__all__ = ()

from scarletio import Compound

from ...channel import Channel
from ...core import GUILDS
from ...guild import Guild
from ...http import DiscordApiClient
from ...payload_building import add_payload_fields_from_keyword_parameters
from ...user import create_partial_user_from_id, ClientUserBase, User, VoiceState
from ...user.guild_profile.utils import GUILD_PROFILE_FIELD_CONVERTERS

from ..request_helpers import (
    get_channel_guild_id_and_id, get_guild_and_id, get_guild_id, get_role_guild_id_and_id, get_user_and_id, get_user_id
)


def _assert__guild_user_search__query(query):
    """
    Asserts the `query` parameter of ``Client.guild_user_search`` method.
    
    Parameters
    ----------
    query : `str`
        The query string with what the user's name or nick should start.
    
    Raises
    ------
    AssertionError
        - If `query` was not given as `str`.
        - If `query`'s length is out of the expected range [1:32].
    """
    if not isinstance(query, str):
        raise TypeError(
            f'`query` can be `str`, got {query.__class__.__name__}; {query!r}'
        )
    
    query_length = len(query)
    if query_length < 1 or query_length > 32:
        raise AssertionError(
            f'`query` length can be in range [1:32], got {query_length!r}; {query!r}.'
        )
    
    return True


def _assert__guild_user_search__limit(limit):
    """
    Asserts the `limit` parameter of ``Client.guild_user_search`` method.
    
    Parameters
    ----------
    limit : `int`
        The maximal amount of users to return. Can be in range [1:1000].
    
    Raises
    ------
    AssertionError
        - If `limit` was not given as `int`.
    """
    if not isinstance(limit, int):
        raise TypeError(
            f'`limit` can be `int`, got {limit.__class__.__name__}; {limit!r}'
        )
    
    return True


class ClientCompoundUserEndpoints(Compound):
    
    api : DiscordApiClient
    id : int
    
    async def user_guild_profile_edit(
        self,
        guild,
        user,
        *,
        reason = None,
        **keyword_parameters,
    ):
        """
        Edits the user's guild profile at the given guild.
        
        This method is a coroutine.
        
        Parameters
        ----------
        guild : ``int | Guild``
            Where the user will be edited.
        
        user : ``ClientUserBase``, `int`
            The user to edit, or their identifier.
        
        reason : `None`, `str` = `None`, Optional (Keyword only)
            Will show up at the guild's audit logs.
        
        **keyword_parameters : Keyword parameters
            Additional keyword parameters representing which field of the guild profile should be edited.
        
        Other Parameters
        ----------------
        deaf : `bool`, Optional (Keyword only)
            Whether the user should be deafen at the voice channels.
        
        mute : `bool`, Optional (Keyword only)
            Whether the user should be muted at the voice channels.
        
        nick : `None`, `str`, Optional (Keyword only)
            The new nick of the user. You can remove the current one by passing it as `None`, an empty string.
        
        voice_channel : ``None | int | Channel`` , Optional (Keyword only)
            Moves the user to the given voice channel. Only applicable if the user is already at a voice channel.
            
            Pass it as `None` to disconnect the user from it's voice channel.
        
        role_ids : `None`, `iterable` of (``Role``, `int`), Optional (Keyword only)
            The new roles of the user. Give it as `None` to remove all of the user's roles.
        
        roles : `None`, `iterable` of (``Role``, `int`), Optional (Keyword only)
            Alternative of `role_ids`.
        
        timed_out_until : `None | DateTime`, Optional (Keyword only)
            Until when the user is timed out.
        
        timeout_duration : `int`, `float`, `TimeDelta`, Optional (Keyword only)
            The timeout duration of the user in seconds or in delta.
        
            The max allowed value equals to 28 days.
        
        voice_channel : ``None | int | Channel`` , Optional (Keyword only)
            Alternative of `voice_channel_id`.
        
        voice_channel_id : ``None | int | Channel`` , Optional (Keyword only)
            Moves the user to the given voice channel. Only applicable if the user is already at a voice channel.
            
            Pass it as `None` to disconnect the user from it's voice channel.
        
        Raises
        ------
        TypeError
            - If a parameter's type is incorrect.
        ValueError
            - If a parameter's type is incorrect.
        ConnectionError
            No internet connection.
        DiscordException
            If any exception was received from the Discord API.
        """
        guild_id = get_guild_id(guild)
        user_id = get_user_id(user)
        
        data = {}
        add_payload_fields_from_keyword_parameters(GUILD_PROFILE_FIELD_CONVERTERS, keyword_parameters, data, True)
        
        if data:
            await self.api.user_guild_profile_edit(guild_id, user_id, data, reason)
    
    
    async def user_role_add(self, user, role, *, reason = None):
        """
        Adds the role on the user.
        
        This method is a coroutine.
        
        Parameters
        ----------
        user : ```ClientUserBase``, `int`
            The user who will get the role.
        role : ``Role | (int, int)``
            The role to add on the user.
        reason : `None`, `str` = `None`, Optional (Keyword only)
            Shows up at the respective guild's audit logs.
        
        Raises
        ------
        TypeError
            - If `user` was not given neither as ``ClientUserBase``, neither as `int`.
            - If `role` was not given neither as ``Role`` nor as `tuple` of (`int`, `int`).
        ConnectionError
            No internet connection.
        DiscordException
            If any exception was received from the Discord API.
        """
        guild_id, role_id = get_role_guild_id_and_id(role)
        user_id = get_user_id(user)
        
        await self.api.user_role_add(guild_id, user_id, role_id, reason)
    
    
    async def user_role_delete(self, user, role, *, reason = None):
        """
        Deletes the role from the user.
        
        This method is a coroutine.
        
        Parameters
        ----------
        user : ```ClientUserBase``, `int`
            The user from who the role will be removed.
        role : ``Role | (int, int)``
            The role to remove from the user.
        reason : `None`, `str` = `None`, Optional (Keyword only)
            Shows up at the respective guild's audit logs.
        
        Raises
        ------
        TypeError
            - If `user` was not given neither as ``ClientUserBase``, neither as `int`.
            - If `role` was not given neither as ``Role`` nor as `tuple` of (`int`, `int`).
        ConnectionError
            No internet connection.
        DiscordException
            If any exception was received from the Discord API.
        """
        guild_id, role_id = get_role_guild_id_and_id(role)
        user_id = get_user_id(user)
        
        await self.api.user_role_delete(guild_id, user_id, role_id, reason)
    
    
    async def user_voice_move(self, user, channel):
        """
        Moves the user to the given voice channel. The user must be in a voice channel at the respective guild already.
        
        This method is a coroutine.
        
        Parameters
        ----------
        user : ```ClientUserBase``, `int`
            The user to move.
        channel : ``Channel``, `tuple` (`int`, `int`)
            The channel where the user will be moved.
        
        Raises
        ------
        TypeError
            - If `user` was not given neither as ``ClientUserBase``, neither as `int`.
            - If `channel` was not given neither as ``Channel`` nor as `tuple` of (`int`, `int`).
        ConnectionError
            No internet connection.
        DiscordException
            If any exception was received from the Discord API.
        """
        guild_id, channel_id = get_channel_guild_id_and_id(channel, Channel.is_in_group_guild_connectable)
        if not guild_id:
            return
        
        user_id = get_user_id(user)
       
        await self.api.user_move(guild_id, user_id, {'channel_id': channel_id})
    
    
    async def user_voice_move_to_speakers(self, user, channel):
        """
        Moves the user to the speakers inside of a stage channel.
        
        This method is a coroutine.
        
        Parameters
        ----------
        user : ```ClientUserBase``, `int`
            The user to move.
        channel : ``Channel``, `tuple` (`int`, `int`)
            The channel where the user will be moved.
        
        Raises
        ------
        TypeError
            - If `user` was not given neither as ``ClientUserBase``, neither as `int`.
            - If `channel` was not given neither as ``Channel`` nor as `tuple` of (`int`, `int`).
        ConnectionError
            No internet connection.
        DiscordException
            If any exception was received from the Discord API.
        """
        guild_id, channel_id = get_channel_guild_id_and_id(channel, Channel.is_guild_stage)
        if not guild_id:
            return
        
        user_id = get_user_id(user)
       
        data = {
            'suppress' : False,
            'channel_id': channel_id,
        }
        
        await self.api.voice_state_edit(guild_id, user_id, data)
    
    
    async def user_voice_move_to_audience(self, user, channel):
        """
        Moves the user to the audience inside of a stage channel.
        
        This method is a coroutine.
        
        Parameters
        ----------
        user : ```ClientUserBase``, `int`
            The user to move.
        channel : ``Channel``, `tuple` (`int`, `int`)
            The channel where the user will be moved.
        
        Raises
        ------
        TypeError
            - If `user` was not given neither as ``ClientUserBase``, neither as `int`.
            - If `channel` was not given neither as ``Channel`` nor as `tuple` of (`int`, `int`).
        ConnectionError
            No internet connection.
        DiscordException
            If any exception was received from the Discord API.
        """
        guild_id, channel_id = get_channel_guild_id_and_id(channel, Channel.is_guild_stage)
        if not guild_id:
            return
        
        user_id = get_user_id(user)
       
        data = {
            'suppress': True,
            'channel_id': channel_id,
        }
        
        await self.api.voice_state_edit(guild_id, user_id, data)
    
    
    async def user_get(self, user, *, force_update = False):
        """
        Gets an user by it's id. If the user is already loaded updates it.
        
        This method is a coroutine.
        
        Parameters
        ----------
        user : ``ClientUserBase``, `int`
            The user, who will be requested.
        force_update : `bool` = `False`, Optional (Keyword only)
            Whether the user should be requested even if it supposed to be up to date.
        
        Returns
        -------
        user : ``ClientUserBase``
        
        Raises
        ------
        TypeError
            If `user` was not given neither as ``ClientUserBase``, neither as `int`.
        ConnectionError
            No internet connection.
        DiscordException
            If any exception was received from the Discord API.
        """
        user, user_id = get_user_and_id(user)
        
        # a goto to check whether we should force update the user.
        while True:
            if force_update:
                break
            
            if user is None:
                break
            
            for guild_id in user.guild_profiles.keys():
                try:
                    guild = GUILDS[guild_id]
                except KeyError:
                    continue
                
                if not guild.partial:
                    return user
            
            break
        
        data = await self.api.user_get(user_id)
        return User.from_data(data)
    
    
    async def guild_user_get(self, guild, user):
        """
        Gets an user and it's profile at a guild. The user must be the member of the guild. If the user is already
        loaded updates it.
        
        This method is a coroutine.
        
        Parameters
        ----------
        guild : ``int | Guild``
            The guild, where the user is.
        user : ```ClientUserBase``, `int`
            The user's id, who will be requested.
        
        Returns
        -------
        user : ``ClientUserBase``
        
        Raises
        ------
        TypeError
            - If `user` was not given neither as ``ClientUserBase``, neither as `int`.
            - If `guild` was not given neither as ``Guild`` nor `int`.
        ConnectionError
            No internet connection.
        DiscordException
            If any exception was received from the Discord API.
        """
        user_id = get_user_id(user)
        guild_id = get_guild_id(guild)
        
        data = await self.api.guild_user_get(guild_id, user_id)
        
        return User.from_data(data['user'], data, guild_id)
    
    
    async def guild_user_search(self, guild, query, limit = 1):
        """
        Gets an user and it's profile at a guild by it's name. If the users are already loaded updates it.
        
        This method is a coroutine.
        
        Parameters
        ----------
        guild : ``Guild``
            The guild, where the user is.
        
        query : `name`
            The query string with what the user's name or nick should start.
        
        limit : `int` = `1`, Optional
            The maximal amount of users to return. Can be in range [1:1000].
            
            When passed as non-positive value, will default to the maximal allowed amount.
        
        Returns
        -------
        users : ``list<ClientUserBase>``
        
        Raises
        ------
        TypeError
            If `guild` was not given neither as ``Guild`` nor `int`.
        ConnectionError
            No internet connection.
        DiscordException
            If any exception was received from the Discord API.
        AssertionError
            - If `limit` was not given as `str`.
            - If `limit` is out fo expected range [1:1000].
        """
        guild_id = get_guild_id(guild)
        
        assert _assert__guild_user_search__query(query)
        assert _assert__guild_user_search__limit(limit)
        
        if limit <= 0 or limit > 1000:
            limit = 1000
        
        data = {'query': query}
        
        if limit != 1:
            data['limit'] = limit
        
        datas = await self.api.guild_user_search(guild_id, data)
        
        return [User.from_data(data['user'], data, guild_id) for data in datas]
    
    
    async def user_voice_kick(self, user, guild):
        """
        Kicks the user from the guild's voice channels. The user must be in a voice channel at the guild.
        
        This method is a coroutine.
        
        Parameters
        ----------
        user : ```ClientUserBase``, `int`
            The user who will be kicked from the voice channel.
        guild : ``int | Guild``
            The guild from what's voice channel the user will be kicked.
        
        Raises
        ------
        TypeError
            - If a parameter's type is incorrect.
        ConnectionError
            No internet connection.
        DiscordException
            If any exception was received from the Discord API.
        """
        user_id = get_user_id(user)
        guild_id = get_guild_id(guild)
        
        await self.api.user_move(guild_id, user_id, {'channel_id': None})
    
    
    async def voice_state_get(self, guild, user, *, force_update = False):
        """
        Gets the user's voice state at the given guild.
                
        This method is a coroutine.
        
        Parameters
        ----------
        guild : ``int | Guild``
            The respective guild.
        user : ```ClientUserBase``, `int`
            The user to get its voice state.
        force_update : `bool` = `False`, Optional (Keyword only)
            Whether the user should be requested even if it supposed to be up to date.
        
        Returns
        -------
        voice_state : ``VoiceState``
        
        Raises
        ------
        TypeError
            - If a parameter's type is incorrect.
        ConnectionError
            No internet connection.
        DiscordException
            If any exception was received from the Discord API.
            
        """
        user, user_id = get_user_and_id(user)
        guild, guild_id = get_guild_and_id(guild)
        
        while True:
            if force_update:
                break
            
            if (guild is None) or guild.partial:
                break
            
            voice_state = guild.get_voice_state(user_id)
            if (voice_state is None):
                break
            
            return voice_state
            
        voice_state_data = await self.api.voice_state_get(guild_id, user_id)
        
        # Try to re-get the guild.
        if guild is None:
            guild = GUILDS.get(guild_id, None)
        
        if guild is None:
            return VoiceState.from_data(voice_state_data, guild_id, strong_cache = False)
        
        if user is None:
            user = create_partial_user_from_id(user_id)
        
        return guild._update_voice_state_restricted(voice_state_data, user)
    
    
    async def voice_state_get_own(self, guild, *, force_update = False):
        """
        Gets the user's voice state at the given guild.
                
        This method is a coroutine.
        
        Parameters
        ----------
        guild : ``int | Guild``
            The respective guild.
        force_update : `bool` = `False`, Optional (Keyword only)
            Whether the user should be requested even if it supposed to be up to date.
        
        Returns
        -------
        voice_state : ``VoiceState``
        
        Raises
        ------
        TypeError
            - If a parameter's type is incorrect.
        ConnectionError
            No internet connection.
        DiscordException
            If any exception was received from the Discord API.
        """
        guild, guild_id = get_guild_and_id(guild)
        
        while True:
            if force_update:
                break
            
            if (guild is None) or guild.partial:
                break
            
            voice_state = guild.get_voice_state(self.id)
            if (voice_state is None):
                break
            
            return voice_state
        
        voice_state_data = await self.api.voice_state_get_own(guild_id)
        
        # Try to re-get the guild.
        if guild is None:
            guild = GUILDS.get(guild_id, None)
        
        if guild is None:
            return VoiceState.from_data(voice_state_data, guild_id, strong_cache = False)
        
        return guild._update_voice_state_restricted(voice_state_data, self)



__all__ = ()

from ...field_parsers import entity_id_parser_factory, nullable_unix_time_parser_factory
from ...field_putters import entity_id_optional_putter_factory, nullable_unix_time_optional_putter_factory
from ...field_validators import entity_id_validator_factory, nullable_date_time_validator_factory

# sku_id

parse_sku_id = entity_id_parser_factory('sku_id')
put_sku_id = entity_id_optional_putter_factory('sku_id')
validate_sku_id = entity_id_validator_factory('sku_id', NotImplemented, include = 'SKU')


# expires_at

parse_expires_at = nullable_unix_time_parser_factory('expires_at')
put_expires_at = nullable_unix_time_optional_putter_factory('expires_at')
validate_expires_at = nullable_date_time_validator_factory('expires_at')



from .ban_add_multiple_result import *
from .ban_entry import *
from .discovery import *
from .discovery_category import *
from .guild import *
from .guild_activity_overview import *
from .guild_activity_overview_activity import *
from .guild_activity_overview_tag import *
from .guild_badge import *
from .guild_boost import *
from .guild_enhancement_entitlements_create_event import *
from .guild_enhancement_entitlements_delete_event import *
from .guild_incidents import *
from .guild_inventory_settings import *
from .guild_join_request import *
from .guild_join_request_delete_event import *
from .guild_join_request_form_response import *
from .guild_preview import *
from .guild_user_chunk_event import *
from .guild_widget import *
from .guild_widget_channel import *
from .guild_widget_user import *
from .verification_screen import *
from .verification_screen_step import *
from .welcome_screen import *
from .welcome_screen_channel import *


__all__ = (
    *ban_add_multiple_result.__all__,
    *ban_entry.__all__,
    *discovery.__all__,
    *discovery_category.__all__,
    *guild.__all__,
    *guild_activity_overview.__all__,
    *guild_activity_overview_activity.__all__,
    *guild_activity_overview_tag.__all__,
    *guild_badge.__all__,
    *guild_boost.__all__,
    *guild_enhancement_entitlements_create_event.__all__,
    *guild_enhancement_entitlements_delete_event.__all__,
    *guild_incidents.__all__,
    *guild_inventory_settings.__all__,
    *guild_join_request.__all__,
    *guild_join_request_delete_event.__all__,
    *guild_join_request_form_response.__all__,
    *guild_preview.__all__,
    *guild_user_chunk_event.__all__,
    *guild_widget.__all__,
    *guild_widget_channel.__all__,
    *guild_widget_user.__all__,
    *verification_screen.__all__,
    *verification_screen_step.__all__,
    *welcome_screen.__all__,
    *welcome_screen_channel.__all__,
)



from .application import *
from .application_executable import *
from .application_install_parameters import *
from .application_integration_type_configuration import *
from .application_role_connection import *
from .application_role_connection_metadata import *
from .client_platform_configuration import *
from .embedded_activity_configuration import *
from .entitlement import *
from .eula import *
from .sku import *
from .sku_enhancement import *
from .sku_enhancement_guild import *
from .subscription import *
from .team import *
from .team_member import *
from .third_party_sku import *


__all__ = (
    *application.__all__,
    *application_executable.__all__,
    *application_install_parameters.__all__,
    *application_role_connection.__all__,
    *application_role_connection_metadata.__all__,
    *client_platform_configuration.__all__,
    *embedded_activity_configuration.__all__,
    *entitlement.__all__,
    *eula.__all__,
    *sku.__all__,
    *sku_enhancement.__all__,
    *sku_enhancement_guild.__all__,
    *subscription.__all__,
    *team.__all__,
    *team_member.__all__,
    *third_party_sku.__all__,
)



__all__ = ()

from ...sticker import StickerFormat, StickerType
from ...sticker.sticker.constants import SORT_VALUE_MIN
from ...sticker.sticker.fields import (
    validate_available, validate_description, validate_format, validate_guild_id, validate_id, validate_name,
    validate_sort_value, validate_tags, validate_type
)

from ..audit_log_entry_change_conversion import AuditLogEntryChangeConversion, AuditLogEntryChangeConversionGroup
from ..audit_log_entry_change_conversion.change_deserializers import change_deserializer_deprecation
from ..conversion_helpers.converters import (
    value_deserializer_description, value_deserializer_id, value_deserializer_name, value_serializer_description, value_serializer_id,
    value_serializer_name
)

# ---- asset ----

ASSET_CONVERSION_IGNORED = AuditLogEntryChangeConversion(
    ('asset',),
    '',
    change_deserializer = change_deserializer_deprecation,
)

# ---- available ----

AVAILABLE_CONVERSION = AuditLogEntryChangeConversion(
    ('available',),
    'available',
    value_validator = validate_available,
)


@AVAILABLE_CONVERSION.set_value_deserializer
def available_value_deserializer(value):
    if value is None:
        value = True
    return value


# ---- description ----

DESCRIPTION_CONVERSION = AuditLogEntryChangeConversion(
    ('description',),
    'description',
    value_deserializer = value_deserializer_description,
    value_serializer = value_serializer_description,
    value_validator = validate_description,
)


# ---- format ----

FORMAT_CONVERSION = AuditLogEntryChangeConversion(
    ('format_type',),
    'format',
    value_validator = validate_format,
)


@FORMAT_CONVERSION.set_value_deserializer
def format_value_deserializer(value):
    return StickerFormat(value)


@FORMAT_CONVERSION.set_value_serializer
def format_value_serializer(value):
    return value.value


# ---- guild_id ----

GUILD_ID_CONVERSION = AuditLogEntryChangeConversion(
    ('guild_id',),
    'guild_id',
    value_deserializer = value_deserializer_id,
    value_serializer = value_serializer_id,
    value_validator = validate_guild_id,
)


# ---- id ----

ID_CONVERSION = AuditLogEntryChangeConversion(
    ('id',),
    'id',
    value_deserializer = value_deserializer_id,
    value_serializer = value_serializer_id,
    value_validator = validate_id,
)


# ---- name ----

NAME_CONVERSION = AuditLogEntryChangeConversion(
    ('name',),
    'name',
    value_deserializer = value_deserializer_name,
    value_serializer = value_serializer_name,
    value_validator = validate_name,
)


# ---- sort_value ----

SORT_VALUE_CONVERSION = AuditLogEntryChangeConversion(
    ('sort_value',),
    'sort_value',
    value_validator = validate_sort_value,
)


@SORT_VALUE_CONVERSION.set_value_deserializer
def sort_value_value_deserializer(value):
    if value is None:
        value = SORT_VALUE_MIN
    return value


# ---- tags ----

TAGS_CONVERSION = AuditLogEntryChangeConversion(
    ('tags',),
    'tags',
    value_validator = validate_tags,
)


@TAGS_CONVERSION.set_value_deserializer
def tags_value_deserializer(value):
    if value is None:
        pass
    elif (not value):
        value = None
    else:
        value = frozenset(value.split(', '))
    
    return value


@TAGS_CONVERSION.set_value_serializer
def tags_value_serializer(value):
    if value is None:
       value = ''
    else:
        value = ', '.join(sorted(value))
    
    return value


# ---- type ----

TYPE_CONVERSION = AuditLogEntryChangeConversion(
    ('type',),
    'type',
    value_validator = validate_type,
)


@TYPE_CONVERSION.set_value_deserializer
def type_value_deserializer(value):
    return StickerType(value)


@TYPE_CONVERSION.set_value_serializer
def type_value_serializer(value):
    return value.value


# ---- Construct ----

STICKER_CONVERSIONS = AuditLogEntryChangeConversionGroup(
    ASSET_CONVERSION_IGNORED,
    AVAILABLE_CONVERSION,
    DESCRIPTION_CONVERSION,
    FORMAT_CONVERSION,
    GUILD_ID_CONVERSION,
    ID_CONVERSION,
    NAME_CONVERSION,
    SORT_VALUE_CONVERSION,
    TAGS_CONVERSION,
    TYPE_CONVERSION,
)



__all__ = ('VerificationScreenStep', )

from scarletio import RichAttributeErrorBaseType

from .fields import (
    parse_required, parse_title, parse_type, parse_values, put_required, put_title, put_type,
    put_values, validate_required, validate_title, validate_type, validate_values
)
from .preinstanced import VerificationScreenStepType


class VerificationScreenStep(RichAttributeErrorBaseType):
    """
    Represents a step of a ``VerificationScreen``.
    
    Attributes
    ----------
    required : `bool`
        Whether the user must accept this step to continue.
    title : `str`
        The step's title.
    type : ``VerificationScreenStepType``
        The type of the step.
    values : `None | tuple<str>`
        The values of the step.
    """
    __slots__ = ('required', 'title', 'type', 'values')
    
    
    def __new__(cls, *, required = ..., step_type = ..., title = ..., values= ...):
        """
        Creates a new verification screen step.
        
        Parameters
        ----------
        required : `bool`, Optional (Keyword only)
            Whether the user must accept this step to continue.
        step_type : ``VerificationScreenStepType``, Optional (Keyword only)
            The type of the step.
        title : `str`, Optional (Keyword only)
            The step's title.
        values : `None`, `iterable` of `str`, Optional (Keyword only)
            The values of the step.
        
        Raises
        ------
        TypeError
            - If a parameter's type is invalid.
        ValueError
            - If a parameter's value is invalid.
        """
        # required
        if required is ...:
            required = False
        else:
            required = validate_required(required)
        
        # title
        if title is ...:
            title = ''
        else:
            title = validate_title(title)
        
        # step_type
        if step_type is ...:
            step_type = VerificationScreenStepType.rules
        else:
            step_type = validate_type(step_type)
        
        # values
        if values is ...:
            values = None
        else:
            values = validate_values(values)
        
        # Construct
        self = object.__new__(cls)
        self.required = required
        self.title = title
        self.type = step_type
        self.values = values
        return self
    
    
    @classmethod
    def from_data(cls, data):
        """
        Creates a new verification screen step from the given data.
        
        Parameters
        ----------
        data : `dict<str, object>`
            Verification screen step data.
        """
        self = object.__new__(cls)
        self.required = parse_required(data)
        self.title = parse_title(data)
        self.type = parse_type(data)
        self.values = parse_values(data)
        return self
    
    
    def to_data(self, *, defaults = False):
        """
        Converts the verification screen step to a json serializable object.
        
        Parameters
        ----------
        defaults : `bool` = `False`, Optional (Keyword only)
            Whether default values should be included as well.
        
        Returns
        -------
        data : `dict<str, object>`
        """
        data = {}
        put_required(self.required, data, defaults)
        put_title(self.title, data, defaults)
        put_type(self.type, data, defaults)
        put_values(self.values, data, defaults)
        return data
    
    
    def __repr__(self):
        """Returns the verification screen step's representation."""
        repr_parts = ['<', self.__class__.__name__]
        
        repr_parts.append(' title = ')
        repr_parts.append(repr(self.title))
        
        step_type = self.type
        repr_parts.append(', type = ')
        repr_parts.append(step_type.name)
        
        required = self.required
        if required:
            repr_parts.append(', required = ')
            repr_parts.append(repr(required))
        
        values = self.values
        if (values is not None):
            repr_parts.append(', values = [')
            
            index = 0
            limit = len(values)
            
            while True:
                value = values[index]
                
                repr_parts.append(repr(value))
                
                index += 1
                if index == limit:
                    break
                
                repr_parts.append(', ')
                continue
            
            repr_parts.append(']')
        
        repr_parts.append('>')
        return ''.join(repr_parts)
    
    
    def __hash__(self):
        """Returns the verification screen step's hash value."""
        hash_value = 0
        
        # required
        hash_value ^= self.required << 14
        
        # title
        hash_value ^= hash(self.title)
        
        # type
        hash_value ^= hash(self.type)
        
        # values
        values = self.values
        if (values is not None):
            hash_value ^= len(values)
            
            for value in values:
                hash_value ^= hash(value)
        
        return hash_value
    
    
    def __eq__(self, other):
        """Returns whether the two verification screen steps are equal"""
        if type(self) is not type(other):
            return NotImplemented
        
        # required
        if self.required != other.required:
            return False
        
        # title
        if self.title != other.title:
            return False
        
        # type
        if self.type is not other.type:
            return False
        
        # values
        if self.values != other.values:
            return False
        
        return True
    
    
    def copy(self):
        """
        Copies the verification screen step.
        
        Returns
        -------
        new : `instance<type<self>>`
        """
        new = object.__new__(type(self))
        new.required = self.required
        new.title = self.title
        new.type = self.type
        values = self.values
        if (values is not None):
            values = (*values,)
        new.values = values
        return new
    
    
    def copy_with(self, *, required = ..., step_type = ..., title = ..., values= ...):
        """
        Copies the verification step with the given fields.
        
        Parameters
        ----------
        required : `bool`, Optional (Keyword only)
            Whether the user must accept this step to continue.
        step_type : ``VerificationScreenStepType``, Optional (Keyword only)
            The type of the step.
        title : `str`, Optional (Keyword only)
            The step's title.
        values : `None`, `iterable` of `str`, Optional (Keyword only)
            The values of the step.
        
        Returns
        -------
        new : `instance<type<self>>`
        
        Raises
        ------
        TypeError
            - If a parameter's type is invalid.
        ValueError
            - If a parameter's value is invalid.
        """
        # required
        if required is ...:
            required = self.required
        else:
            required = validate_required(required)
        
        # title
        if title is ...:
            title = self.title
        else:
            title = validate_title(title)
        
        # step_type
        if step_type is ...:
            step_type = self.type
        else:
            step_type = validate_type(step_type)
        
        # values
        if values is ...:
            values = self.values
            if (values is not None):
                values = (*values,)
        else:
            values = validate_values(values)
        
        # Construct
        new = object.__new__(type(self))
        new.required = required
        new.title = title
        new.type = step_type
        new.values = values
        return new
    
    
    def iter_values(self):
        """
        Iterates over the values of the verification screen.
        
        This method is an iterable generator.
        
        Yields
        ------
        value : `str`
        """
        values = self.values
        if (values is not None):
            yield from values



from .application_command import *
from .application_command_autocomplete import *
from .base import *
from .constants import *
from .fields import *
from .form_submit import *
from .message_component import *


__all__ = (
    *application_command.__all__,
    *application_command_autocomplete.__all__,
    *base.__all__,
    *constants.__all__,
    *fields.__all__,
    *form_submit.__all__,
    *message_component.__all__,
)



__all__ = ()

from ...field_parsers import bool_parser_factory, nullable_date_time_parser_factory
from ...field_putters import force_bool_putter_factory, nullable_date_time_optional_putter_factory
from ...field_validators import (
    bool_validator_factory, force_date_time_validator_factory, nullable_date_time_validator_factory
)
from ...utils import DISCORD_EPOCH_START, datetime_to_id, id_to_datetime


# cancelled

parse_cancelled = bool_parser_factory('is_canceled', False)
put_cancelled = force_bool_putter_factory('is_canceled')
validate_cancelled = bool_validator_factory('cancelled', False)


# end

parse_end = nullable_date_time_parser_factory('scheduled_end_time')
put_end = nullable_date_time_optional_putter_factory('scheduled_end_time')
validate_end = nullable_date_time_validator_factory('end')


# start

parse_start = nullable_date_time_parser_factory('scheduled_start_time')
put_start = nullable_date_time_optional_putter_factory('scheduled_start_time')
validate_start = nullable_date_time_validator_factory('start')


# timestamp

def parse_timestamp(data):
    """
    Parses cancellation timestamp.
    
    Parameters
    ----------
    data : `dict<str, object>`
        Data to parse from.
    
    Returns
    -------
    timestamp : `DateTime`
    """
    timestamp_as_id = data.get('event_exception_id', None)
    if timestamp_as_id is None:
        return DISCORD_EPOCH_START
    
    return id_to_datetime(int(timestamp_as_id))


def put_timestamp(timestamp, data, defaults):
    """
    Serialises the timestamp into the given data.
    
    Parameters
    ----------
    timestamp : `None | tuple<DateTime>`
        The timestamp to serialize.
    
    data : `dict<str, object>`
        Json serializable dictionary.
    
    defaults : `bool`
        Whether default values should be included as well.
    
    Returns
    -------
    data : `dict<str, object>`
    """
    data['event_exception_id'] = datetime_to_id(timestamp)
    return data


validate_timestamp = force_date_time_validator_factory('timestamp')



from .client_user_base import *
from .client_user_presence_base import *
from .constants import *
from .fields import *
from .flags import *
from .matching import *
from .helpers import *
from .orin_user_base import *
from .preinstanced import *
from .user import *
from .user_base import *
from .utils import *


__all__ = (
    *client_user_base.__all__,
    *client_user_presence_base.__all__,
    *constants.__all__,
    *fields.__all__,
    *flags.__all__,
    *helpers.__all__,
    *matching.__all__,
    *orin_user_base.__all__,
    *preinstanced.__all__,
    *user.__all__,
    *user_base.__all__,
    *utils.__all__,
)



__all__ = ()

import sys
from os import sep as PATH_SEPARATOR
from os.path import (
    basename as get_file_name, commonpath as _get_common_path, isabs as is_absolute_path, split as split_path
)
from sys import path as system_paths

from scarletio import any_to_any, get_short_executable

from ... import __package__ as PACKAGE_NAME

from .constants import PACKAGE_MAIN_FILE, UPPER_DIRECTORY, WORKING_DIRECTORY


# Conditional imports
if sys.platform == 'linux':
    from os import geteuid as get_effective_user_id
else:
    get_effective_user_id = None


def get_main_call(with_parameters = False):
    """
    Returns how the library main was called.
    
    Parameters
    ----------
    with_parameters : `bool` = `False`, Optional
        Whether call parameters should also be rendered.
    
    Returns
    -------
    main_call : `str`
    """
    return ''.join(render_main_call_into([], with_parameters = with_parameters))


def _render_default_main_call_into(into):
    """
    Renders the the default call way into the given list.
    
    Parameters
    ----------
    into : `list` of `str`
        List to extend.
    
    Returns
    -------
    into : `list` of `str`
    """
    into.append(get_short_executable())
    into.append(' -m ')
    into.append(PACKAGE_NAME)
    return into


def _render_parameters_into(into, parameters):
    """
    Parameters
    ----------
    into : `list` of `str`
        List to extend.
    parameters : `list` of `str`
        The parameters to render.
    
    Returns
    -------
    into : `list` of `str`
    """
    for parameter in parameters:
        into.append(' ')
        if not parameter:
            into.append('""')
            continue
        
        if not any_to_any(parameter, (' ', '"')):
            into.append(parameter)
            continue
        
        into.append('"')
        into.append(parameter.replace('"', '\\"'))
        into.append('"')
        continue

    return into


def get_common_path(paths):
    """
    Returns the longest common path.
    
    Parameters
    ----------
    paths : `iterable<str>`
        Paths.
    
    Returns
    -------
    sub_path : `str`
    """
    try:
        return _get_common_path(paths)
    except ValueError as e:
        # Ignore if the paths are on a different drive lol.
        if e.args and e.args[0] == 'Paths don\'t have the same drive':
            return ''
        
        raise


def normalize_executed_file(executed_file):
    """
    Normalizes the executed file.
    
    Parameters
    ----------
    executed_file : `str`
        File name or path to normalize.
    
    Returns
    -------
    executed_file : `str`
        The executed file normalized.
    library_mode : `bool`
        Whether the executed file was probably ran as a library.
    """
    if not is_absolute_path(executed_file):
        return executed_file, False
    
    directory_name, file_name = split_path(executed_file)
    if file_name == '__main__.py':
        executed_file = directory_name
    
    longest_common_path = max(
        (get_common_path((executed_file, system_path),) for system_path in system_paths),
        default = None,
        key = (lambda path: len(path)),
    )
    if longest_common_path is None:
        return executed_file, True
    
    executed_file = executed_file[len(longest_common_path):]
    if executed_file.startswith(PATH_SEPARATOR):
        executed_file = executed_file[len(PATH_SEPARATOR):]
    
    if not executed_file:
        executed_file = '.'
    
    return executed_file, True


def render_main_call_into(into, with_parameters = False):
    """
    Renders how the library was called into the given list.
    
    Parameters
    ----------
    into : `list` of `str`
        List to extend.
    with_parameters : `bool` = `False`, Optional
        Whether call parameters should also be rendered.
    
    Returns
    -------
    into : `list` of `str`
    """
    if (get_effective_user_id is not None) and (get_effective_user_id() == 0):
        into.append('sudo ')
    
    system_parameters = sys.argv
    
    while True:
        if len(system_parameters) < 1:
            into = _render_default_main_call_into(into)
            with_parameters = False
            break
        
        executed_file = system_parameters[0]
        if executed_file == PACKAGE_MAIN_FILE:
            into = _render_default_main_call_into(into)
            break
            
        if (UPPER_DIRECTORY != WORKING_DIRECTORY):
            executed_file_name = get_file_name(executed_file)
            if'.' not in executed_file_name:
                into.append(executed_file_name)
                break
        
        into.append(get_short_executable())
        into.append(' ')
        executed_file, library_mode = normalize_executed_file(executed_file)
        if library_mode:
            into.append('-m ')
        into.append(executed_file)
        break
        
    if with_parameters:
        into = _render_parameters_into(into, system_parameters[1:])
    
    return into



__all__ = ()

from os.path import  dirname as get_directory_name, join as join_paths
from scarletio import get_frame_module, get_last_module_frame
from scarletio.tools.asynchronous_interactive_console import (
    collect_module_variables, create_banner, create_exit_message, run_asynchronous_interactive_console
)
from scarletio.tools.asynchronous_interactive_console.console_helpers import AWAIT_NOTE

from .... import __package__ as PACKAGE_NAME

from ...core import register


PACKAGE = __import__(PACKAGE_NAME)


def is_module_library_main(module):
    """
    Returns whether the module is the library's main file.
    
    Parameters
    ----------
    module : `ModuleType`
        The module to analyze.
    
    Returns
    -------
    is_module_library_main : `bool`
    """
    module_file_name = getattr(module, '__file__', None)
    if module_file_name is None:
        return False
    
    package_file_name = getattr(PACKAGE, '__file__', None)  
    if (package_file_name is None):
        return False
    
    if module_file_name != join_paths(get_directory_name(package_file_name), '__main__.py'):
        return False
    
    return True


def collect_variables():
    """
    Collects variables for the asynchronous console.
    
    Returns
    -------
    module_variables : `dict<str, object>`
    """
    module = get_frame_module(get_last_module_frame())
    if (module is None) or is_module_library_main(module):
        return collect_module_variables(PACKAGE)
    
    return collect_module_variables(module)


@register(
    aliases = 'i',
    description =  f'Runs asynchronous python interpreter through scarletio.\n{AWAIT_NOTE}',
)
def interpreter():
    """
    Runs asynchronous console.
    """
    run_asynchronous_interactive_console(
        collect_variables(),
        banner = create_banner(PACKAGE),
        exit_message = create_exit_message(PACKAGE),
    )



__all__ = ('merge_update_reaction_mapping',)

from ..reaction_mapping_line import merge_update_reaction_mapping_lines


def merge_update_reaction_mapping(new_reactions, old_reactions):
    """
    Merges the two reaction mapping, so values wont be overwritten if not required.
    
    Parameters
    ----------
    new_reactions : `None`, ``ReactionMapping``
        The new reactions on a message.
    old_reactions : `None`, ``ReactionMapping``
        The old reactions on a message.
    
    Returns
    -------
    real_reactions : `None`, ``ReactionMapping``
        The real merged reactions on a message.
    """
    if (old_reactions is None):
        return new_reactions
    
    if (new_reactions is None):
        old_reactions.clear()
        return old_reactions
    
    old_reactions.lines = merge_update_reaction_mapping_lines(new_reactions.lines, old_reactions.lines)
    return old_reactions



__all__ = ()


# ---- channel ----

def _channel_match_sort_key(item):
    """
    Sort key used inside of ``Guild.get_channels_like`` to sort channels based on their match rate.
    
    Parameters
    ----------
    item : `tuple` (``Channel``, `tuple` (`int`, `int`))
        The channel and it's match rate.
    
    Returns
    -------
    match_rate : `tuple` (`int`, `int`)
    """
    return item[1]

# ---- emoji ----

def _emoji_match_sort_key(item):
    """
    Sort key used inside of ``Guild.get_emojis_like`` to sort emojis based on their match rate.
    
    Parameters
    ----------
    item : `tuple` (``Emoji``, `tuple` (`int`, `int`))
        The emoji and it's match rate.
    
    Returns
    -------
    match_rate : `tuple` (`int`, `int`)
    """
    return item[1]


def _strip_emoji_name(name):
    """
    Strips the given emoji name from colon signs.
    
    Parameters
    ----------
    name : `str`
        The name to strip.
    
    Returns
    -------
    name : `str`
    """
    if name.startswith(':'):
        starts_at = 1
    else:
        starts_at = None
    
    if name.endswith(':'):
        ends_at = -1
    else:
        ends_at = None
    
    if (starts_at is not None) or (ends_at is not None):
        name = name[starts_at : ends_at]
    
    return name


# ---- role ----
def _role_match_sort_key(item):
    """
    Sort key used inside of ``Guild.get_roles_like`` to sort roles based on their match rate.
    
    Parameters
    ----------
    item : `tuple` (``Role | (int, int)``)
        The role and it's match rate.
    
    Returns
    -------
    match_rate : `tuple` (`int`, `int`)
    """
    return item[1]


# ---- soundboard sound ----

def _soundboard_sound_match_sort_key(item):
    """
    Sort key used inside of ``Guild.get_soundboard_sounds_like`` to sort soundboard sounds based on their match rate.
    
    Parameters
    ----------
    item : `tuple` (``Sticker``, `tuple` (`int`, `int`))
        The soundboard sound and it's match rate.
    
    Returns
    -------
    match_rate : `tuple` (`int`, `int`)
    """
    return item[1]


# ---- sticker ----

STICKER_MATCH_WEIGHT_NAME = 1
STICKER_MATCH_WEIGHT_TAG = 2


def _sticker_match_sort_key(item):
    """
    Sort key used inside of ``Guild.get_stickers_like`` to sort stickers based on their match rate.
    
    Parameters
    ----------
    item : `tuple` (``Sticker``, `tuple` (`int`, `int`, `int`))
        The sticker and it's match rate.
    
    Returns
    -------
    match_rate : `tuple` (`int`, `int`, `int`)
    """
    return item[1]



__all__ = ()

from ...field_parsers import (
    default_entity_parser_factory, entity_id_parser_factory, nullable_date_time_parser_factory,
    nullable_entity_parser_factory, nullable_object_array_parser_factory, nullable_string_parser_factory,
    preinstanced_parser_factory
)
from ...field_putters import (
    entity_id_putter_factory, entity_putter_factory, nullable_date_time_optional_putter_factory,
    nullable_entity_optional_putter_factory, nullable_object_array_optional_putter_factory, preinstanced_putter_factory,
    url_optional_putter_factory
)
from ...field_validators import (
    default_entity_validator_factory, entity_id_validator_factory, nullable_date_time_validator_factory,
    nullable_entity_validator_factory, nullable_object_array_validator_factory, nullable_string_validator_factory,
    preinstanced_validator_factory
)
from ...user import ClientUserBase, User, ZEROUSER

from ..guild import Guild
from ..guild_join_request_form_response import GuildJoinRequestFormResponse

from .preinstanced import GuildJoinRequestStatus

# actioned_at

parse_actioned_at = nullable_date_time_parser_factory('actioned_at')
put_actioned_at = nullable_date_time_optional_putter_factory('actioned_at')
validate_actioned_at = nullable_date_time_validator_factory('actioned_at')

# actioned_by

parse_actioned_by = nullable_entity_parser_factory('actioned_by_user', User)
put_actioned_by = nullable_entity_optional_putter_factory('actioned_by_user', User, force_include_internals = True)
validate_actioned_by = nullable_entity_validator_factory('actioned_by', ClientUserBase)

# created_at

parse_created_at = nullable_date_time_parser_factory('created_at')
put_created_at = nullable_date_time_optional_putter_factory('created_at')
validate_created_at = nullable_date_time_validator_factory('created_at')

# form_responses

parse_form_responses = nullable_object_array_parser_factory('form_responses', GuildJoinRequestFormResponse)
put_form_responses = nullable_object_array_optional_putter_factory('form_responses')
validate_form_responses = nullable_object_array_validator_factory('form_responses', GuildJoinRequestFormResponse)

# guild_id

parse_guild_id = entity_id_parser_factory('guild_id')
put_guild_id = entity_id_putter_factory('guild_id')
validate_guild_id = entity_id_validator_factory('guild_id', Guild)

# last_seen_at

parse_last_seen_at = nullable_date_time_parser_factory('last_seen')
put_last_seen_at = nullable_date_time_optional_putter_factory('last_seen')
validate_last_seen_at = nullable_date_time_validator_factory('last_seen')

# rejection_reason

parse_rejection_reason = nullable_string_parser_factory('rejection_reason')
put_rejection_reason = url_optional_putter_factory('rejection_reason')
validate_rejection_reason = nullable_string_validator_factory('rejection_reason', 0, 1024)

# status

parse_status = preinstanced_parser_factory('application_status', GuildJoinRequestStatus, GuildJoinRequestStatus.started)
put_status = preinstanced_putter_factory('application_status')
validate_status = preinstanced_validator_factory('application_status', GuildJoinRequestStatus)

# user

parse_user = default_entity_parser_factory('user', User, default = ZEROUSER)
put_user = entity_putter_factory('user', ClientUserBase, force_include_internals = True)
validate_user = default_entity_validator_factory('user', ClientUserBase, default = ZEROUSER)

# user_id

put_user_id = entity_id_putter_factory('user_id')



from .application_command_permission_overwrite import *
from .fields import *
from .helpers import *
from .preinstanced import *


__all__ = (
    *application_command_permission_overwrite.__all__,
    *fields.__all__,
    *helpers.__all__,
    *preinstanced.__all__,
)



__all__ = ('get_bool_env', 'get_int_env', 'get_str_env',)

from os import getenv as get_environmental_variable
from warnings import warn


ERROR_MESSAGE_APPENDIX = 'Please update your `.env` file.'


RETURN_TYPE_VALUE = 0
RETURN_TYPE_WARNING = 1
RETURN_TYPE_EXCEPTION = 2


def _process_bool_env(env_variable):
    """
    Processes the given `bool` environmental variable.
    
    Parameters
    ----------
    env_variable : `str`
        The variable to process.
    
    Returns
    -------
    accepted : `bool`
    value : `bool`
    """
    env_variable = env_variable.casefold()
    if env_variable in ('true', '1'):
        accepted = True
        value = True
    
    elif env_variable in ('false', '0'):
        accepted = True
        value = False
    
    else:
        accepted = False
        value = False
    
    return accepted, value


def _process_str_env(env_variable):
    """
    Processes a `str` environmental variable.
    
    Parameters
    ----------
    env_variable : `str`
        The variable to process.
    
    Returns
    -------
    accepted : `bool`
    value : `str`
    """
    return (True if env_variable else False), env_variable


def _process_int_env(env_variable):
    """
    Processes a `int` environmental variable.
    
    Parameters
    ----------
    env_variable : `str`
        The variable to process.
    
    Returns
    -------
    accepted : `bool`
    value : `int`
    """
    try:
        variable = int(env_variable)
    except ValueError:
        accepted = False
        variable = 0
        
    else:
        accepted = True
    
    return accepted, variable


def _get_env(name, default, accepted_type_name, accepted_processor, raise_if_missing_or_empty, warn_if_empty):
    """
    Gets the environmental variable for the given name. Processes and validates it.
    
    This function is an iterable generator.
    
    Parameters
    ----------
    name : `str`
        The name of an environmental variable.
    default : `object`
        Default value to return.
    accepted_type_name : `str`
        The accepted type's name.
    accepted_type_processor : `(str) -> (bool, object)`
        Processor to process the environmental variable's value.
    raise_if_missing_or_empty : `bool`
        Whether exception should be thrown if the environmental variable is missing or empty.
    warn_if_empty : `bool`
        Whether warning should be dropped if empty environmental variable is received.
    
    Returns
    -------
    return_type : `int`
    return_value / error_message : `object` / `str`
        Returns the returns' type and its value.
    """
    env_variable = get_environmental_variable(name)
    if env_variable is None:
        if raise_if_missing_or_empty:
            yield RETURN_TYPE_EXCEPTION, f'Environmental variable {name!r} is missing. {ERROR_MESSAGE_APPENDIX}'
            return
        
        yield RETURN_TYPE_VALUE, default
        return
    
    accepted, value = accepted_processor(env_variable)
    if accepted:
        yield RETURN_TYPE_VALUE, value
        return
    
    if env_variable:
        yield (
            RETURN_TYPE_WARNING,
            f'{name!r} is specified as non {accepted_type_name}: {env_variable!r}, defaulting to {default!r}!'
        )
    else:
        if raise_if_missing_or_empty:
            yield (
                RETURN_TYPE_EXCEPTION,
                f'Environmental variable {name!r} is specified as empty string. {ERROR_MESSAGE_APPENDIX}'
            )
            return
        
        if warn_if_empty:
            yield (
                RETURN_TYPE_WARNING,
                f'Environmental variable {name!r} is specified as empty string. Defaulting to {default!r}!',
            )
    
    yield RETURN_TYPE_VALUE, default


def _handle_get_env_generator(generator):
    """
    Handles a ``_get_env`` generator.
    
    Parameters
    ----------
    generator : `iterable<(int, object)>`
        Generator to iterate over.
    
    Returns
    -------
    value : `object`
    
    Raises
    ------
    RuntimeError
    """
    for return_type, return_value in generator:
        if return_type == RETURN_TYPE_VALUE:
            return return_value
        
        if return_type == RETURN_TYPE_WARNING:
            warn(return_value, stacklevel = 3)
            continue
        
        if return_type == RETURN_TYPE_EXCEPTION:
            raise RuntimeError(return_value)


def get_bool_env(name, default = False, *, raise_if_missing_or_empty = False, warn_if_empty = True):
    """
    Gets the given environmental variable.
    
    If the environmental variable is not present or not present as a boolean's representation returns `default` instead.
    
    Parameters
    ----------
    name : `str`
        The name of an environmental variable.
    default : `bool` = `False`, Optional
        The default value of the respective variable.
    raise_if_missing_or_empty : `bool` = `False`, Optional (Keyword only)
        Whether exception should be thrown if the environmental variable is missing or empty.
        Has priority over `default` and `warn_if_empty`.
    warn_if_empty : `bool` = `True`, Optional (Keyword only)
        Whether warning should be dropped if empty environmental variable is received.
    
    Returns
    -------
    variable : `bool`
    
    Raises
    ------
    RuntimeError
    """
    return _handle_get_env_generator(
        _get_env(name, default, 'bool', _process_bool_env, raise_if_missing_or_empty, warn_if_empty)
    )


def get_str_env(name, default = None, *, raise_if_missing_or_empty = False, warn_if_empty = True):
    """
    Gets the given environmental variable.
    
    If the environmental variable is not present or present as an empty string returns `default` instead.
    
    Parameters
    ----------
    name : `str`
        The name of an environmental variable.
    default : `None` | `str` = `None`, Optional
        The default value of the respective variable. Defaults to `None`.
    raise_if_missing_or_empty : `bool` = `False`, Optional (Keyword only)
        Whether exception should be thrown if the environmental variable is missing or empty.
        Has priority over `default` and `warn_if_empty`.
    warn_if_empty : `bool` = `True`, Optional (Keyword only)
        Whether warning should be dropped if empty environmental variable is received.
    
    Returns
    -------
    variable : `None` | `str`
    
    Raises
    ------
    RuntimeError
    """
    return _handle_get_env_generator(
        _get_env(name, default, 'str', _process_str_env, raise_if_missing_or_empty, warn_if_empty)
    )


def get_int_env(name, default = 0, *, raise_if_missing_or_empty = False, warn_if_empty = True):
    """
    Gets the given environmental variable.
    
    If the environmental variable is not present or present as non `int`, will return `default` instead.
    
    Parameters
    ----------
    name : `str`
        The name of an environmental variable.
    default : `int` = `0`, Optional
        The default value of the respective variable.
    raise_if_missing_or_empty : `bool` = `False`, Optional (Keyword only)
        Whether exception should be thrown if the environmental variable is missing or empty.
        Has priority over `default` and `warn_if_empty`.
    warn_if_empty : `bool` = `True`, Optional (Keyword only)
        Whether warning should be dropped if empty environmental variable is received.
    
    Returns
    -------
    variable : `int`
    
    Raises
    ------
    RuntimeError
    """
    return _handle_get_env_generator(
        _get_env(name, default, 'int', _process_int_env, raise_if_missing_or_empty, warn_if_empty)
    )



__all__ = ()

from ...scheduled_event.scheduled_event_occasion_overwrite.fields import (
    validate_cancelled, validate_end, validate_start
)
from ...utils import datetime_to_timestamp, timestamp_to_datetime

from ..audit_log_entry_change_conversion import AuditLogEntryChangeConversion, AuditLogEntryChangeConversionGroup


# ---- cancelled ----

CANCELLED_CONVERSION = AuditLogEntryChangeConversion(
    ('is_canceled',),
    'cancelled',
    value_validator = validate_cancelled,
)


@CANCELLED_CONVERSION.set_value_deserializer
def cancelled_value_deserializer(value):
    if value is None:
        value = False
    return value


# ---- end ----

END_CONVERSION = AuditLogEntryChangeConversion(
    ('scheduled_end_time',),
    'end',
    value_validator = validate_end,
)


# ---- start ----

START_CONVERSION = AuditLogEntryChangeConversion(
    ('scheduled_start_time',),
    'start',
    value_validator = validate_start,
)

# ----

@END_CONVERSION.set_value_deserializer
@START_CONVERSION.set_value_deserializer
def date_time_converter_get(value):
    if (value is not None):
        value = timestamp_to_datetime(value)
    return value

@END_CONVERSION.set_value_serializer
@START_CONVERSION.set_value_serializer
def date_time_converter_put(value):
    if (value is not None):
        value = datetime_to_timestamp(value)
    return value


# ---- Construct ----

SCHEDULED_EVENT_OCCASION_OVERWRITE_CONVERSIONS = AuditLogEntryChangeConversionGroup(
    CANCELLED_CONVERSION,
    END_CONVERSION,
    START_CONVERSION,
)



from .api_client import *
from .connector_cache import *
from .headers import *
from .rate_limit import *
from .rate_limit_groups import *
from .rate_limit_proxy import *
from .urls import *

from . import rate_limit_groups as RATE_LIMIT_GROUPS


__all__ = (
    'RATE_LIMIT_GROUPS',
    
    *api_client.__all__,
    *connector_cache.__all__,
    *headers.__all__,
    *rate_limit.__all__,
    *rate_limit_groups.__all__,
    *rate_limit_proxy.__all__,
    *urls.__all__,
)



__all__ = ()

from scarletio import include


from ...field_parsers import (
    bool_parser_factory, default_entity_parser_factory, entity_id_parser_factory, int_parser_factory,
    nullable_date_time_parser_factory, nullable_entity_parser_factory, preinstanced_array_parser_factory,
    preinstanced_parser_factory
)
from ...field_putters import (
    bool_optional_putter_factory, entity_id_optional_putter_factory, entity_putter_factory, int_optional_putter_factory,
    nullable_date_time_optional_putter_factory, nullable_entity_putter_factory, preinstanced_array_putter_factory,
    preinstanced_putter_factory
)
from ...field_validators import (
    bool_validator_factory, default_entity_validator_factory, entity_id_validator_factory, entity_validator_factory,
    int_conditional_validator_factory, int_options_validator_factory, nullable_date_time_validator_factory,
    nullable_entity_validator_factory, preinstanced_array_validator_factory, preinstanced_validator_factory
)
from ...oauth2 import Oauth2Scope
from ...role import Role
from ...user import ClientUserBase, User

from ..integration_account import IntegrationAccount
from ..integration_application import IntegrationApplication

from .constants import EXPIRE_GRACE_PERIOD_DEFAULT, EXPIRE_GRACE_PERIOD_OPTIONS, SUBSCRIBER_COUNT_DEFAULT
from .preinstanced import IntegrationExpireBehavior


IntegrationType = include('IntegrationType')

# account

parse_account = default_entity_parser_factory(
    'account', IntegrationAccount, default_factory = lambda : IntegrationAccount._create_empty()
)
put_account = entity_putter_factory('account', IntegrationAccount)
validate_account = default_entity_validator_factory(
    'account', IntegrationAccount, default_factory = lambda : IntegrationAccount._create_empty()
)

# account [discord]

def parse_account__discord(data):
    """
    Parses out an integration account of a Discord integration out from the given data.
    
    Parameters
    ----------
    data : `dict<str, object>`
        Integration account data.
    
    Returns
    -------
    integration_account : ``ClientUserBase``
    """
    account_data = data.get('account', None)
    if (account_data is None):
        return IntegrationAccount._create_empty()
    
    user_id = int(account_data['id'])
    name = account_data['name']
    
    return User.precreate(user_id, name = name, bot = True)


def put_account__discord(account, data, defaults):
    """
    Serialises the given discord integration account into the given `data`.
    
    Parameters
    ----------
    account : ``ClientUserBase``
        Integration account of a discord integration.
    data : `dict<str, object>`
        Json serializable dictionary.
    defaults : `bool`
        Whether default fields should be included as well.
    
    Returns
    -------
    data : `dict<str, object>`
    """
    data['account'] = {
        'id': str(account.id),
        'name': account.name,
    }
    
    return data

validate_account__discord = entity_validator_factory('account', ClientUserBase)

# application

parse_application = nullable_entity_parser_factory('application', IntegrationApplication)
put_application = nullable_entity_putter_factory('application', IntegrationApplication)
validate_application = nullable_entity_validator_factory('application', IntegrationApplication)

# scopes

parse_scopes = preinstanced_array_parser_factory('scopes', Oauth2Scope)
put_scopes = preinstanced_array_putter_factory('scopes')
validate_scopes = preinstanced_array_validator_factory('scopes', Oauth2Scope)

# validate_emojis_enabled

validate_emojis_enabled = bool_validator_factory('emojis_enabled', True)

# expire_behavior

parse_expire_behavior = preinstanced_parser_factory(
    'expire_behavior', IntegrationExpireBehavior, IntegrationExpireBehavior.remove_role,
)
put_expire_behavior = preinstanced_putter_factory('expire_behavior')
validate_expire_behavior = preinstanced_validator_factory('expire_behavior', IntegrationExpireBehavior)

# expire_grace_period

parse_expire_grace_period = int_parser_factory('expire_grace_period', EXPIRE_GRACE_PERIOD_DEFAULT)
put_expire_grace_period = int_optional_putter_factory('expire_grace_period', EXPIRE_GRACE_PERIOD_DEFAULT)
validate_expire_grace_period = int_options_validator_factory('expire_grace_period', EXPIRE_GRACE_PERIOD_OPTIONS, 0)

# revoked

parse_revoked = bool_parser_factory('revoked', False)
put_revoked = bool_optional_putter_factory('revoked', False)
validate_revoked = bool_validator_factory('revoked', False)

# role_id

parse_role_id = entity_id_parser_factory('role_id')
put_role_id = entity_id_optional_putter_factory('role_id')
validate_role_id = entity_id_validator_factory('role_id', Role)

# subscriber_count

parse_subscriber_count = int_parser_factory('subscriber_count', SUBSCRIBER_COUNT_DEFAULT)
put_subscriber_count = int_optional_putter_factory('subscriber_count', SUBSCRIBER_COUNT_DEFAULT)
validate_subscriber_count = int_conditional_validator_factory(
    'subscriber_count', 0, lambda integer: integer >= SUBSCRIBER_COUNT_DEFAULT, 'positive'
)

# synced_at

parse_synced_at = nullable_date_time_parser_factory('synced_at')
put_synced_at = nullable_date_time_optional_putter_factory('synced_at')
validate_synced_at = nullable_date_time_validator_factory('synced_at')

# syncing

parse_syncing = bool_parser_factory('syncing', False)
put_syncing = bool_optional_putter_factory('syncing', False)
validate_syncing = bool_validator_factory('syncing', False)



__all__ = ()

from .plugin_tree import PluginTree


def _is_plugin_in_plugin_trees(plugin, plugin_trees):
    """
    Returns whether the plugin is in an of the given plugin trees.
    
    Parameters
    ----------
    plugin : ``Plugin``
        The plugin to check its presence of.
    plugin_trees : `set` of ``PluginTree``
        Already built plugin trees.
    
    Returns
    -------
    is_plugin_in_plugin_trees : `bool`
    """
    for plugin_tree in plugin_trees:
        if plugin_tree.has_plugin(plugin):
            return True
    
    return False


def _remove_subset_plugin_trees(plugin_tree, plugin_trees):
    """
    Removes subset only plugins trees. Their entry point must match too.
    
    Parameters
    ----------
    plugin_tree : ``PluginTree``
        The plugin tree to check against.
    plugin_trees : `set` of ``PluginTree``
        Already built plugin trees.
    """
    plugin_trees_to_remove = None
    
    for plugin_tree_to_check in plugin_trees:
        if plugin_tree.get_entry_plugin() is not plugin_tree_to_check.get_entry_plugin():
            continue
        
        if not plugin_tree_to_check.is_subset_of_plugin_tree(plugin_tree):
            continue
        
        if plugin_trees_to_remove is None:
            plugin_trees_to_remove = []
        
        plugin_trees_to_remove.append(plugin_tree_to_check)
        
    if (plugin_trees_to_remove is not None):
        plugin_trees.discard(plugin_tree)


def _build_plugin_trees(plugins, plugin_trees, deep):
    """
    Builds plugin trees from the given plugins.
    
    Parameters
    ----------
    plugins : `iterable` of ``Plugin``
        Plugins to build trees for.
    plugin_trees : `iterable` of ``PluginTree``
        Already matched plugin trees if any.
    deep : `bool`
        Whether the plugin with all of it's parent and with their child should be returned.
    
    Returns
    -------
    plugin_trees : `set` of ``PluginTree``
    """
    plugin_trees = {*plugin_trees}
    
    for plugin in plugins:
        if _is_plugin_in_plugin_trees(plugin, plugin_trees):
            continue
        
        plugin_tree = PluginTree(plugin, deep)
        _remove_subset_plugin_trees(plugin_tree, plugin_trees)
        plugin_trees.add(plugin_tree)
    
    return plugin_trees


def _sort_plugin_trees(plugin_trees):
    """
    Sorts the given plugin trees.
    
    Parameters
    ----------
    plugin_trees : `set` of ``PluginTree``
        Plugin trees to sort.
    
    Returns
    -------
    plugin_trees : `list` of ``PluginTree``
    """
    return sorted(plugin_trees)


def _build_and_sort_plugin_trees(plugins, plugin_trees, deep):
    """
    Builds plugin tree and sorts them.
    
    Parameters
    ----------
    plugins : `iterable` of ``Plugin``
        A list of plugin to build the tree form.
    plugin_trees : `iterable` of ``PluginTree``
        Already matched plugin trees if any.
    deep : `bool`
        Whether the plugin with all of it's parent and with their child should be returned.
    
    Returns
    -------
    plugin_trees : `list` of ``PluginTree``
    """
    return _sort_plugin_trees(_build_plugin_trees(plugins, plugin_trees, deep))


def _unwrap_plugin_trees_into_plugins(plugin_trees):
    """
    Unwraps the plugin trees.
    
    Parameters
    ----------
    plugin_trees : `list` of ``PluginTree``
        Plugin trees to unwrap.
    
    Returns
    -------
    plugins : `list` of ``PluginTree``
    """
    plugins = []
    
    for plugin_tree in plugin_trees:
        plugins.extend(plugin_tree.get_plugins_sorted())
    
    return plugins



__all__ = ()

from scarletio import RichAttributeErrorBaseType

from ....discord.events.handling_helpers import _EventHandlerManager, Router


class NestableInterface(RichAttributeErrorBaseType):
    """
    Common class for nestable instances.
    """
    __slots__ = ()
    
    
    def _is_nestable(self):
        """
        Returns whether the instance can be nested.
        
        Returns
        -------
        nestable : `bool`.
        """
        return True
    
    
    def _check_supports_nesting(self):
        """
        If the instance is not nestable raises `RuntimeError`.
        
        Raises
        ------
        RuntimeError
            The instance is not nestable.
        """
        if not self._is_nestable():
            raise RuntimeError(f'{self!r} is not nestable.')
    
    
    @property
    def interactions(self):
        """
        Enables you to add sub-commands or sub-categories to the command.
        
        Raises
        ------
        RuntimeError
            The instance is not nestable.
        """
        self._check_supports_nesting()
        return _EventHandlerManager(self)
    
    
    def create_event(self, function, *positional_parameters, **keyword_parameters):
        """
        Adds a sub-command under the slash command.
        
        Parameters
        ----------
        func : `async-callable`
            The function used as the command when using the respective slash command.
        
        *positional_parameters : Positional Parameters
            Positional parameters to pass to ``SlashCommand``'s constructor.
        
        **keyword_parameters : Keyword parameters
            Keyword parameters to pass to the ``SlashCommand``'s constructor.
        
        Returns
        -------
        instance : `object`
        
        Raises
        ------
        TypeError
            - If a parameter's type is incorrect.
        ValueError
            - If a parameter's value is incorrect.
        RuntimeError
            - If cant register sub-command.
        """
        self._check_supports_nesting()
        
        if isinstance(function, Router):
            function = function[0]
        
        added, instance = self._store_command_instance(function)
        if added:
            return instance
        
        command = self._make_command_instance_from_parameters(function, positional_parameters, keyword_parameters)
        
        added, instance = self._store_command_instance(command)
        if added:
            return instance
        
        return command
    
    
    def _make_command_instance_from_parameters(self, function, positional_parameters, keyword_parameters):
        """
        Creates a command instance from the given parameters.
        
        Parameters
        ----------
        function : `None | CoroutineFunctionType | CoroutineGeneratorFunctionType`
            Function to create command from.
        positional_parameters : `tuple<object>`
            Captured positional parameters.
        keyword_parameters : `dict<str, object>`
            Captured keyword parameters.
        
        Returns
        -------
        instance : `object`
        
        Raises
        ------
        TypeError
            - If a parameter's type is incorrect.
        ValueError
            - If a parameter's value is incorrect.
        RuntimeError
            - If cant register sub-command.
        """
        return None
    
    
    def _store_command_instance(self, command):
        """
        Registers a nestable instance.
        
        Parameters
        ----------
        command : `object`
            Command instance to register.
        
        Returns
        -------
        registered : `bool`
        instance : `None | object`
        """
        return False, None



__all__ = ()

from itertools import chain


def _unwrap_plugin(plugin, deep):
    """
    Unwraps the given plugin.
    
    Parameters
    ----------
    plugin : ``Plugin``
        The plugin to unwrap.
    deep : `bool`
        Whether the plugin with all of it's parent and with their child should be returned.
    
    Returns
    -------
    plugins : `set` of ``Plugin``
    """
    plugins_to_unwrap = [plugin]
    unwrapped_plugins = set()
    
    while plugins_to_unwrap:
        plugin = plugins_to_unwrap.pop()
        
        if deep:
            for iterated_plugin in chain(
                plugin.iter_child_plugins(),
                plugin.iter_parent_plugins(),
                plugin.iter_sub_module_plugins()
            ):
                if iterated_plugin not in unwrapped_plugins:
                    plugins_to_unwrap.append(iterated_plugin)
        
        unwrapped_plugins.add(plugin)
    
    return unwrapped_plugins


def _sort_plugins(plugins):
    """
    Sorts the given plugins.
    
    Parameters
    ----------
    plugins : `set` of ``Plugin``
        Plugin to sort.
    
    Returns
    -------
    plugins : `list` of ``Plugin``
        Sorted plugins.
    """
    plugins_to_check_ordered = sorted(plugins, reverse = True)
    
    plugins_satisfied = set()
    plugins_satisfied_ordered = []
    
    while plugins_to_check_ordered:
        for index in reversed(range(len(plugins_to_check_ordered))):
            plugin = plugins_to_check_ordered[index]
            
            if not plugin.are_child_plugins_present_in(plugins_satisfied):
                continue
            
            plugins_satisfied.add(plugin)
            plugins_satisfied_ordered.append(plugin)
            del plugins_to_check_ordered[index]
            break

        else:
            raise RuntimeError(
                f'Plugins with circular satisfaction: {plugins_to_check_ordered!r}; plugin = {plugin!r}.'
            )
    
    return plugins_satisfied_ordered



__all__ = ()

from scarletio import include

from ..core import APPLICATION_ID_TO_CLIENT

from .intent import INTENT_MASK_DIRECT_MESSAGES, INTENT_MASK_GUILD_MESSAGES, INTENT_MASK_MESSAGE_CONTENT


Client = include('Client')


def filter_clients(iterator, flag_mask, me):
    """
    Filters the clients whether their intents allows the specific flag.
    
    First yields the first client from `clients` what allows the specified flag. If non, then yields `None`.
    If a `None` or not the expected client was yielded, then the generator should be closed.
    
    If the correct client was yielded, then the generator is used at a for loop yielding all the clients from `clients`
    which allow the specified flag including the firstly yielded one.
    
    This function is a generator.
    
    Parameters
    ----------
    iterator : ``GeneratorType<Client>``
        An iterator over the clients to filter from.
    
    flag_mask : `int`
        The intent flag's mask based on what the clients will be filtered.
    me : ``Client``
        The client who received the respective event.
    
    Yields
    -------
    client : ``Client``, `None`
    """
    for client in iterator:
        if client.intents & flag_mask == flag_mask:
            break
    
    else:
        yield me
        yield me
        return
    
    yield client
    yield client
    
    for client in iterator:
        if client.intents & flag_mask == flag_mask:
            yield client



def get_message_enabled_user_ids(message_data):
    """
    Gets the content field enabled user id-s for the given message.
    
    Parameters
    ----------
    message_data : `dict<str, object>`
        Received message data.
    
    Returns
    -------
    enabled_user_ids : `set` of `int`
    """
    enabled_user_ids = set()
    
    try:
        author_data = message_data['author']
    except KeyError:
        application_id = message_data.get('application_id')
        if (application_id is not None):
            application_id = int(application_id)
            
            try:
                client = APPLICATION_ID_TO_CLIENT[application_id]
            except KeyError:
                pass
            else:
                enabled_user_ids.add(client.id)
            
    else:
        user_id = int(author_data['id'])
        enabled_user_ids.add(user_id)
    
    try:
        user_mention_datas = message_data['mentions']
    except KeyError:
        pass
    else:
        for user_mention_data in user_mention_datas:
            user_id = int(user_mention_data['id'])
            enabled_user_ids.add(user_id)
    
    return enabled_user_ids


def filter_content_intent_client(iterator, message_data, me):
    """
    Filters the clients who has message content intent.
    
    First yields the first client from `clients` that allows message content. If non, then yields `None`.
    If a `None` or not the expected client was yielded, then the generator should be closed.
    
    If the correct client was yielded, then the generator is used at a for loop yielding all the clients from `clients`
    which allow the specified flag including the firstly yielded one.
    
    This function is a generator.
    
    Parameters
    ----------
    iterator : ``GeneratorType<Client>``
        An iterator over the client to filter from.
    
    message_data : `dict<str, object>`
        Received message data.
    
    me : ``Client``
        The client who received the respective event.
    
    Yields
    -------
    client : ``Client``
    """
    # Check whether any of the clients has the required intent mask
    flag_mask = INTENT_MASK_MESSAGE_CONTENT
    
    if message_data.get('guild_id', None) is None:
        flag_mask |= INTENT_MASK_DIRECT_MESSAGES
    else:
        flag_mask |= INTENT_MASK_GUILD_MESSAGES
    
    enabled_user_ids = get_message_enabled_user_ids(message_data)
    
    for client in iterator:
        if (client.intents & flag_mask == flag_mask) or (client.id in enabled_user_ids):
            break
    
    else:
        yield me
        yield me
        return
    
    yield client
    yield client
    
    for client in iterator:
        if (client.intents & flag_mask == flag_mask) or (client.id in enabled_user_ids):
            yield client


def filter_clients_or_me(iterator, flag_mask, me):
    """
    Filters the clients whether their intents allow the specific flag. This filter is used, when the clients receive
    the respective event for themselves even if they have the intent disabled.
    
    First yields the first client from `clients` what allows the specified flag.
    
    If non of the clients allow it, then yields `me` (so the source client), what received the event, then expects
    a `user` to be yielded back. At the end when the generator is iterated over inside of a for loop, then it yields
    `me` again (expect if `me` is not the same as the back yielded `user`, but that should not happen, but making sure).
    
    If any of the source clients allow the specified intent flag, then yields the first client what allows it.
    If not the correct client was yielded back, then the generator should be closed. Meanwhile if the correct client
    was yielded, then the generator expects a `user` to be yielded back. After it, the generator is used inside
    of a for loop yielding all the clients from `clients` which allow the specified intent flag including the firstly
    yielded one. At the end yields the received `user` if it is type ``Client`` and it's specified intent flag is not
    allowed.
    
    This function is a generator.
    
    Parameters
    ----------
    iterator : ``GeneratorType<Client>``
        An iterator over the clients to filter from.
    
    flag_mask : `int`
        The intent flag's mask based on what the clients will be filtered.
    
    me : ``Client``
        The client who received the respective event.
    
    Yields
    -------
    client : ``Client``, `None`
    """
    for client in iterator:
        if client.intents & flag_mask == flag_mask:
            break
    
    else:
        user = yield me
        yield
        if user is me:
            yield me
        return
    
    user = yield client
    yield
    yield client
    
    for client in iterator:
        if client.intents & flag_mask == flag_mask:
            yield client
    
    # Whether the user is type Client and we did not yield it, yield it.
    if not isinstance(user, Client):
        return
    
    if user.intents & flag_mask == flag_mask:
        return
    
    yield user


def filter_just_me(me):
    """
    Yields the source client.
    
    This filter is used when the client itself is the only one who should handle the respective event.
    
    Parameters
    ----------
    me : ``Client``
        The client to yield back.
    
    Yields
    -------
    client : ``Client``
    """
    yield me


def first_client(iterator, flag_mask, me):
    """
    Returns the first client what allows the specified intent flag. If no client allows it, then returns `None`.
    
    Parameters
    ----------
    iterator : ``GeneratorType<Client>``
        An iterator over the clients to filter from.
    
    flag_mask : `int`
        The intent flag's mask based on what the clients will be filtered.
    
    me : ``Client``
        The client who received the respective event.
    
    Returns
    -------
    client : ``Client``
    """
    for client in iterator:
        if client.intents & flag_mask == flag_mask:
            return client
    
    return me


def first_content_intent_client(iterator, message_data, me):
    """
    Returns the first client, who has message content intent for the given message.
    
    If no client allows it, returns the first client.
    
    Parameters
    ----------
    iterator : ``GeneratorType<Client>``
        An iterator over the clients to filter from.
    
    message_data : `dict<str, object>`
        Received message data.
    
    me : ``Client``
        The client who received the respective event.
    
    Returns
    -------
    client : ``Client``
    """
    clients = [*iterator]
    # Fast check for obvious speed reasons.
    if (len(clients) <= 1) or (message_data.get('flags', 0) & (1 << 6)):
        return me
    
    # Check whether any of the clients has the required intent mask
    flag_mask = INTENT_MASK_MESSAGE_CONTENT
    if message_data.get('guild_id', None) is None:
        flag_mask |= INTENT_MASK_DIRECT_MESSAGES
    else:
        flag_mask |= INTENT_MASK_GUILD_MESSAGES
    
    for client in clients:
        if client.intents & flag_mask == flag_mask:
            return client
    
    enabled_user_ids = get_message_enabled_user_ids(message_data)
    for client in clients:
        if client.id in enabled_user_ids:
            return client
    
    return clients[0]


def first_client_or_me(iterator, flag_mask, me):
    """
    Returns the first client what allows the specified intent flag. If non of the clients allow it, then returns `me`.
    
    Parameters
    ----------
    iterator : ``GeneratorType<Client>``
        An iterator over the clients to filter from.
    
    flag_mask : `int`
        The intent flag's mask based on what the clients will be filtered.
    me : ``Client``
        The client who received the respective event.
    
    Returns
    -------
    client : ``Client``
    """
    for client in iterator:
        if client.intents & flag_mask == flag_mask:
            return client
    
    return me



__all__ = ()

from scarletio import include, include_with_callback

from ...field_parsers import (
    field_parser_factory, nullable_array_parser_factory, nullable_entity_parser_factory,
    nullable_object_array_parser_factory
)
from ...field_putters import (
    nullable_entity_optional_putter_factory, nullable_object_array_optional_putter_factory,
    nullable_string_array_optional_putter_factory, nullable_string_optional_putter_factory
)
from ...field_validators import (
    nullable_entity_validator_factory, nullable_object_array_validator_factory,
    nullable_string_array_validator_factory, nullable_string_validator_factory
)

from ..shared_fields import parse_custom_id, put_custom_id, validate_custom_id

# components

InteractionComponent = include('InteractionComponent')


def validate_components_factory(nesting_target_name, any_required_flag_names):
    """
    Creates a new component validator.
    
    Parameters
    ----------
    nesting_target_name : `int`
        The name under what these components will be nested to.
    
    any_required_flag_names : `tuple<str>`
        Required layout flags. Keys to pass as `True` to ``ComponentTypeLayoutFlag.update_by_keys``.
        
        Note that only non-strict matching is supported by the generated function.
    
    Returns
    -------
    validator : ``(None | iterable<InteractionComponent>) -> None | tuple<InteractionComponent>``
    """
    COMPONENT_LAYOUT_MASK_ANY_REQUIRED = 0
    InteractionComponent = NotImplemented
    
    def validator(components):
        """
        Validates the given components and whether they can be nested into a certain other component.
        
        This function is auto generated.
        
        Parameters
        ----------
        components : ``None | iterable<InteractionComponent | (tuple | list)<InteractionComponent>>``
            The components to validate.
        
        Returns
        -------
        components_processed : ``None | tuple<InteractionComponent>``
        
        Raises
        ------
        TypeError
            - If `components` is invalid type.
        ValueError
            - If any of the `components` cannot be nested.
        """
        nonlocal nesting_target_name
        nonlocal COMPONENT_LAYOUT_MASK_ANY_REQUIRED
        nonlocal InteractionComponent
        
        if components is None:
            return None
        
        if (getattr(components, '__iter__', None) is None):
            raise TypeError(
                f'`components` can be `None`, `iterable` of `{InteractionComponent.__name__}`, got '
                f'{type(components).__name__}; {components!r}.'
            )
        
        components_processed = None
        
        for component in components:
            if isinstance(component, InteractionComponent):
                flags = component.type.layout_flags
                if (flags & COMPONENT_LAYOUT_MASK_ANY_REQUIRED):
                    # Allow any component satisfying the flag.
                    pass
                    
                else:
                    raise ValueError(
                        f'Cannot nest components of type {component.type.name} into a {nesting_target_name!s}'
                        f', got {component!r}; components = {components!r}.'
                    )
            
            else:
                raise TypeError(
                    f'`components` can contain `{InteractionComponent.__name__}` elements, got '
                    f'{type(component).__name__}; {component!r}; components = {components!r}.'
                )
            
            if (components_processed is None):
                components_processed = []
            
            components_processed.append(component)
        
        if (components_processed is not None):
            components_processed = tuple(components_processed)
        
        return components_processed
    
    
    @include_with_callback('ComponentTypeLayoutFlag')
    def include_component_type_layout_flag(resolved_ComponentTypeLayoutFlag):
        nonlocal any_required_flag_names
        nonlocal COMPONENT_LAYOUT_MASK_ANY_REQUIRED
        
        COMPONENT_LAYOUT_MASK_ANY_REQUIRED = resolved_ComponentTypeLayoutFlag().update_by_keys(
            **dict.fromkeys(any_required_flag_names, True)
        )
    
    
    @include_with_callback('InteractionComponent')
    def include_component(resolved_Component):
        nonlocal InteractionComponent
        
        InteractionComponent = resolved_Component
    
    
    return validator


parse_components = nullable_object_array_parser_factory('components', NotImplemented, include = 'InteractionComponent')
put_components = nullable_object_array_optional_putter_factory('components')

validate_components = nullable_object_array_validator_factory(
    'components', NotImplemented, include = 'InteractionComponent'
)

validate_components__row = validate_components_factory(
    'row',
    (
        'nestable_into_row',
    ),
)

validate_components__section = validate_components_factory(
    'section',
    (
        'nestable_into_section',
    ),
)

validate_components__container = validate_components_factory(
    'container',
    (
        'nestable_into_container',
    ),
)


# component

parse_component = nullable_entity_parser_factory('component', NotImplemented, include = 'InteractionComponent')
put_component = nullable_entity_optional_putter_factory('component', NotImplemented, can_include_internals = False)


def validate_component__label(component):
    """
    Validates whether the given label component is correct type.
    
    Parameters
    ----------
    component : ``None | InteractionComponent``
        The component to validate.
    
    Returns
    -------
    component : ``None | InteractionComponent``
    
    Raises
    ------
    TypeError
        - If `component` is given as an invalid type.
    """
    if component is None:
        return None
    
    if isinstance(component, InteractionComponent):
        if component.type.layout_flags.nestable_into_label:
            return component
        
        raise ValueError(
            f'Component of type: {component.type.name} cannot be used as a label component, got {component!r}.'
        )
    
    raise TypeError(
        f'`component` can be `{InteractionComponent.__name__}`, got {type(component).__name__}, {component!r}.'
    )


# thumbnail

parse_thumbnail = nullable_entity_parser_factory('thumbnail', NotImplemented, include = 'InteractionComponent')
put_thumbnail = nullable_entity_optional_putter_factory('thumbnail', NotImplemented, can_include_internals = False)


def validate_thumbnail(thumbnail):
    """
    Validates whether the given section thumbnail is correct type.
    
    Parameters
    ----------
    thumbnail : ``None | InteractionComponent``
        The thumbnail to validate.
    
    Returns
    -------
    thumbnail : ``None | InteractionComponent``
    
    Raises
    ------
    TypeError
        - If `thumbnail` is given as an invalid type.
    """
    if thumbnail is None:
        return None
    
    if isinstance(thumbnail, InteractionComponent):
        if thumbnail.type.layout_flags.section_thumbnail:
            return thumbnail
        
        raise ValueError(
            f'Component of type: {thumbnail.type.name} cannot be used as a section thumbnail, got {thumbnail!r}.'
        )
    
    raise TypeError(
        f'`thumbnail` can be `{InteractionComponent.__name__}`, got {type(thumbnail).__name__}, {thumbnail!r}.'
    )

# value

parse_value = field_parser_factory('value')
put_value = nullable_string_optional_putter_factory('value')
validate_value = nullable_string_validator_factory('value', 0, 16384)


# values

parse_values = nullable_array_parser_factory('values')
put_values = nullable_string_array_optional_putter_factory('values')
validate_values = nullable_string_array_validator_factory('values')



__all__ = ()

from ...field_validators import int_conditional_validator_factory
from ...user import ClientUserBase

# count

validate_count = int_conditional_validator_factory(
    'count',
    0,
    (lambda count : count >= 0),
    '>= 0',
)

# users

def validate_users(users):
    """
    Validates the input users of a poll result.
    
    Parameters
    ----------
    users : `None`, `iterable` of ``ClientUserBase``
        The users to validate.
    
    Returns
    -------
    users : ``None | set<ClientUserBase>``
    
    Raises
    ------
    TypeError
        - `users`'s type is invalid.
    """
    if users is None:
        return None
    
    if (getattr(users, '__iter__', None) is None):
        raise TypeError(
            f'`users` can be either `None`  or `iterable` of `{ClientUserBase.__name__}` elements, got {users!r}.'
        )
    
    validated_users = None
    
    
    for user in users:
        if not isinstance(user, ClientUserBase):
            raise TypeError(
                f'`users` can have `{ClientUserBase.__name__}` elements, got '
                f'{type(user).__name__}; {user!r}; users = {users!r}.'
            )
        
        if validated_users is None:
            validated_users = set()
        
        validated_users.add(user)
        continue
        
    
    return validated_users



__all__ = ('Oauth2Access', )

from datetime import datetime as DateTime, timedelta as TimeDelta, timezone as TimeZone
from time import time as time_now

from scarletio import RichAttributeErrorBaseType

from .fields import (
    parse_access_token, parse_expires_after, parse_refresh_token, parse_scopes, put_access_token,
    put_expires_after, put_refresh_token, put_scopes, validate_access_token, validate_created_at,
    validate_expires_after, validate_redirect_url, validate_refresh_token, validate_scopes
)
from .preinstanced import Oauth2Scope


class Oauth2Access(RichAttributeErrorBaseType):
    """
    Represents a Discord oauth2 access object, what is returned by ``Client.activate_authorization_code`` if
    activating the authorization code went successfully.
    
    Attributes
    ----------
    access_token : `str`
        Token used for `Bearer` authorizations, when requesting OAuth2 data about the respective user.
    
    created_at : `DateTime`
        The time when the access was last created or renewed.
    
    expires_after : `int`
        The time in seconds after this access expires.
    
    redirect_url : `str`
        The redirect url with what the user granted the authorization code for the oauth2 scopes for the application.
        
        Can be empty string if application's owner's access was requested.
    
    refresh_token : `str`
        The token used to renew the access token.
        
        Can be empty string if application's owner's access was requested.
    
    scopes : `None`, `tuple` of ``Oauth2Scope``
        A sequence of the scopes, which the user granted with the access token.
        
        Defaults to `None` if empty.
    """
    __slots__ = ('access_token', 'created_at', 'expires_after', 'redirect_url', 'refresh_token', 'scopes',)
    
    def __new__(
        cls,
        *,
        access_token = ...,
        created_at = ...,
        expires_after = ...,
        redirect_url = ...,
        refresh_token = ...,
        scopes = ...,
    ):
        """
        Creates a new oauth2 access with the given fields.
        
        Parameters
        ----------
        access_token : `str`, Optional (Keyword only)
            Token used for `Bearer` authorizations, when requesting OAuth2 data about the respective user.
        
        created_at : `DateTime`, Optional (Keyword only)
            The time when the access was last created or renewed.
        
        expires_after : `int`, Optional (Keyword only)
            The time in seconds after this access expires.
        
        redirect_url : `str`, Optional (Keyword only)
            The redirect url with what the user granted the authorization code for the oauth2 scopes for the application.
        
        refresh_token : `str`, Optional (Keyword only)
            The token used to renew the access token.
        
        scopes : `None`, `iterable` of (`str`, ``Oauth2Scope``), Optional (Keyword only)
            A sequence of the scopes, which the user granted with the access token.
        
        Raises
        ------
        TypeError
            - If a parameter's type is incorrect.
        ValueError
            - If a parameter's value is incorrect.
        """
        # access_token
        if access_token is ...:
            access_token = ''
        else:
            access_token = validate_access_token(access_token)
        
        # created_at
        if created_at is ...:
            created_at = DateTime.now(TimeZone.utc)
        else:
            created_at = validate_created_at(created_at)
        
        # expires_after
        if expires_after is ...:
            expires_after = 604800
        else:
            expires_after = validate_expires_after(expires_after)
        
        # redirect_url
        if redirect_url is ...:
            redirect_url = ''
        else:
            redirect_url = validate_redirect_url(redirect_url)
        
        # refresh_token
        if refresh_token is ...:
            refresh_token = ''
        else:
            refresh_token = validate_refresh_token(refresh_token)
        
        # scopes
        if scopes is ...:
            scopes = None
        else:
            scopes = validate_scopes(scopes)
        
        # Construct
        self = object.__new__(cls)
        self.access_token = access_token
        self.created_at = created_at
        self.expires_after = expires_after
        self.redirect_url = redirect_url
        self.refresh_token = refresh_token
        self.scopes = scopes
        return self
    
    
    @classmethod
    def from_data(cls, data, redirect_url):
        """
        Creates an ``Oauth2Access``.
        
        Parameters
        ----------
        data : `dict<str, object>`
            Received access data.
        redirect_url : `str`
            The redirect url with what the user granted the authorization code for the oauth2 scopes for the
            application.
        
        Returns
        -------
        self : `instance<cls>`
        """
        self = object.__new__(cls)
        self.access_token = parse_access_token(data)
        self.created_at = DateTime.now(TimeZone.utc)
        self.expires_after = parse_expires_after(data)
        self.redirect_url = redirect_url
        self.refresh_token = parse_refresh_token(data)
        self.scopes = parse_scopes(data)
        
        return self
    
    
    def to_data(self, *, defaults = False):
        """
        Converts the access to a json serializable object.
        
        Parameters
        ----------
        defaults : `bool` = `False`, Optional (Keyword only)
            Whether fields with their default value should be included as well.
        
        Returns
        -------
        data : `dict<str, object>`
        """
        data = {}
        put_access_token(self.access_token, data, defaults)
        put_expires_after(self.expires_after, data, defaults)
        put_refresh_token(self.refresh_token, data, defaults)
        put_scopes(self.scopes, data, defaults)
        return data
    
    
    def copy(self):
        """
        Copies the oauth2 access.
        
        Returns
        -------
        new : `instance<type<self>``
        """
        new = object.__new__(type(self))
        new.access_token = self.access_token
        new.created_at = self.created_at
        new.expires_after = self.expires_after
        new.redirect_url = self.redirect_url
        new.refresh_token = self.refresh_token
        scopes = self.scopes
        if (scopes is not None):
            scopes = (*scopes,)
        new.scopes = scopes
        return new
        
    
    def copy_with(
        self,
        *,
        access_token = ...,
        created_at = ...,
        expires_after = ...,
        redirect_url = ...,
        refresh_token = ...,
        scopes = ...,
    ):
        """
        Copies the oauth2 scopes with the given fields.
        
        Parameters
        ----------
        access_token : `str`, Optional (Keyword only)
            Token used for `Bearer` authorizations, when requesting OAuth2 data about the respective user.
        
        created_at : `DateTime`, Optional (Keyword only)
            The time when the access was last created or renewed.
        
        expires_after : `int`, Optional (Keyword only)
            The time in seconds after this access expires.
        
        redirect_url : `str`, Optional (Keyword only)
            The redirect url with what the user granted the authorization code for the oauth2 scopes for the application.
        
        refresh_token : `str`, Optional (Keyword only)
            The token used to renew the access token.
        
        scopes : `None`, `iterable` of (`str`, ``Oauth2Scope``), Optional (Keyword only)
            A sequence of the scopes, which the user granted with the access token.
        
        Returns
        -------
        new : `instance<type<self>``
        
        Raises
        ------
        TypeError
            - If a parameter's type is incorrect.
        ValueError
            - If a parameter's value is incorrect.
        """
        # access_token
        if access_token is ...:
            access_token = self.access_token
        else:
            access_token = validate_access_token(access_token)
        
        # created_at
        if created_at is ...:
            created_at = self.created_at
        else:
            created_at = validate_created_at(created_at)
        
        # expires_after
        if expires_after is ...:
            expires_after = self.expires_after
        else:
            expires_after = validate_expires_after(expires_after)
        
        # redirect_url
        if redirect_url is ...:
            redirect_url = self.redirect_url
        else:
            redirect_url = validate_redirect_url(redirect_url)
        
        # refresh_token
        if refresh_token is ...:
            refresh_token = self.refresh_token
        else:
            refresh_token = validate_refresh_token(refresh_token)
        
        # scopes
        if scopes is ...:
            scopes = self.scopes
            if (scopes is not None):
                scopes = (*scopes,)
        else:
            scopes = validate_scopes(scopes)
        
        # Construct
        new = object.__new__(type(self))
        new.access_token = access_token
        new.created_at = created_at
        new.expires_after = expires_after
        new.redirect_url = redirect_url
        new.refresh_token = refresh_token
        new.scopes = scopes
        return new
    
    
    def _renew(self, data):
        """
        Renews the access with the given data.
        
        Parameters
        ----------
        data : `None | dict<str, object>`
            Requested access data.
        """
        self.created_at = DateTime.now(TimeZone.utc)
        if data is None:
            return
        
        self.access_token = parse_access_token(data)
        self.expires_after = parse_expires_after(data)
        self.refresh_token = parse_refresh_token(data)
        self.scopes = parse_scopes(data)
    
    
    @property
    def expires_at(self):
        """
        Returns when the access expires.
        
        Returns
        -------
        expires_at : `DateTime`
        """
        return self.created_at + TimeDelta(seconds = self.expires_after)
    
    
    def __repr__(self):
        """Returns the representation of the oath2 access."""
        repr_parts = ['<', self.__class__.__name__]
        
        if self.created_at.timestamp() + self.expires_after > time_now():
            state = 'active'
        else:
            state = 'expired'
        
        repr_parts.append(' ')
        repr_parts.append(state)
        
        repr_parts.append('scope count: ')
        repr_parts.append(str(len(self.scopes)))
        
        repr_parts.append('>')
        return ''.join(repr_parts)
    
    
    def __hash__(self):
        """Returns the oauth2 scope's hash value."""
        hash_value = 0
        
        # access_token
        hash_value ^= hash(self.access_token)
        
        # created_at
        hash_value ^= hash(self.created_at)
        
        # expires_after
        hash_value ^= hash(self.expires_after)
        
        # redirect_url
        hash_value ^= hash(self.redirect_url)
        
        # refresh_token
        hash_value ^= hash(self.refresh_token)
        
        # scopes
        scopes = self.scopes
        if (scopes is not None):
            hash_value ^= len(scopes) << 22
            
            for scope in scopes:
                hash_value ^= hash(scope)
        
        return hash_value
    
    
    def __eq__(self, other):
        """Returns whether the two oauth2 access are equal."""
        if type(self) is not type(other):
            return False
        
        # access_token
        if self.access_token != other.access_token:
            return False
        
        # created_at
        if self.created_at != other.created_at:
            return False
        
        # expires_after
        if self.expires_after != other.expires_after:
            return False
        
        # redirect_url
        if self.redirect_url != other.redirect_url:
            return False
        
        # refresh_token
        if self.refresh_token != other.refresh_token:
            return False
        
        # scopes
        if self.scopes != other.scopes:
            return False
        
        return True
    
    
    def has_scope(self, scope):
        """
        Returns whether the access has the given scope.
        
        Parameters
        ----------
        scope : ``Oauth2Scope``, `str`
            The scope to check out.
        
        Returns
        -------
        has_scope : `bool`
        
        Raises
        ------
        TypeError
            - If `scope` is neither `str`, ``Oauth2Scope``.
        """
        if isinstance(scope, Oauth2Scope):
            pass
        
        elif isinstance(scope, str):
            scope = Oauth2Scope(scope)
        
        else:
            raise TypeError(
                f'`scope` can be `str`, `{Oauth2Scope.__name__}`, got {type(scope).__name__}; {scope!r}.'
            )
        
        scopes = self.scopes
        if scopes is None:
            return False
        
        return (scope in scopes)
    
    
    def iter_scopes(self):
        """
        Iterates over the oauth2 scopes of the oauth2 access.
        
        This method is an iterable generator.
        
        Yields
        ------
        scope : ``Oauth2Scope``
        """
        scopes = self.scopes
        if (scopes is not None):
            yield from scopes



__all__ = ()

from scarletio import RichAttributeErrorBaseType


class ParameterConverterBase(RichAttributeErrorBaseType):
    """
    Base class for parameter converters.
    
    Attributes
    ----------
    parameter_name : `str`
        The parameter's name.
    """
    __slots__ = ('parameter_name',)
    
    def __new__(cls, parameter_name):
        """
        Creates a new parameter converter from the given parameter.
        
        Parameters
        ----------
        parameter_name : `str`
            The parameter's name.
        """
        self = object.__new__(cls)
        self.parameter_name = parameter_name
        return self
    
    
    async def __call__(self, client, interaction_event, value):
        """
        Calls the parameter converter to convert the given `value` to it's desired state.
        
        This method is a coroutine.
        
        Parameters
        ----------
        client : ``Client``
            The client who received the respective ``InteractionEvent``.
        
        interaction_event : ``InteractionEvent``
            The received application command interaction.
        
        value : `object`
            Value to convert.
        
        Returns
        -------
        converted_value : `None | object`
            If conversion fails always returns `None`.
        
        Raises
        ------
        SlashCommandParameterConversionError
            The parameter cannot be parsed.
        """
    
    
    def __repr__(self):
        """Returns the parameter converter's representation."""
        repr_parts = ['<', type(self).__name__]
        
        # parameter_name
        repr_parts.append(' parameter_name = ')
        repr_parts.append(repr(self.parameter_name))
        
        self._put_repr_parts_into(repr_parts)
        
        repr_parts.append('>')
        return ''.join(repr_parts)
    
    
    def _put_repr_parts_into(self, repr_parts):
        """
        Helper function for sub-types to extend the created representation.
        
        Parameters
        ----------
        repr_parts : `list<str>`
            Representation parts to extended.
        """
    
    
    def as_option(self):
        """
        Converts the parameter to an application command option if applicable.
        
        Returns
        -------
        option : `None | ApplicationCommandOption`
        """
    
    
    def bind_parent(self, parent):
        """
        Binds the parent command to self.
        
        This method might be called for a few types of command functions to bind themselves to a few interactive
        parameters.
        
        Parameters
        ----------
        parent : `None | SlashCommandFunction`
            The slasher application command function to bind to self.
        """



__all__ = ()

from ...field_parsers import (
    entity_id_parser_factory, force_string_parser_factory, nullable_unix_time_parser_factory,
    preinstanced_parser_factory
)
from ...field_putters import (
    entity_id_optional_putter_factory, force_string_putter_factory, nullable_unix_time_optional_putter_factory,
    preinstanced_putter_factory
)
from ...field_validators import (
    entity_id_validator_factory, force_string_validator_factory, nullable_date_time_validator_factory,
    preinstanced_validator_factory
)

from .preinstanced import Palette

# asset_path

parse_asset_path = force_string_parser_factory('asset')
put_asset_path = force_string_putter_factory('asset')
validate_asset_path = force_string_validator_factory('asset_path', 0, 1024)


# expires_at

parse_expires_at = nullable_unix_time_parser_factory('expires_at')
put_expires_at = nullable_unix_time_optional_putter_factory('expires_at')
validate_expires_at = nullable_date_time_validator_factory('expires_at')


# name

parse_name = force_string_parser_factory('label')
put_name = force_string_putter_factory('label')
validate_name = force_string_validator_factory('name', 0, 1024)


# palette

parse_palette = preinstanced_parser_factory('palette', Palette, Palette.black)
put_palette = preinstanced_putter_factory('palette')
validate_palette = preinstanced_validator_factory('palette', Palette)


# sku_id

parse_sku_id = entity_id_parser_factory('sku_id')
put_sku_id = entity_id_optional_putter_factory('sku_id')
validate_sku_id = entity_id_validator_factory('sku_id', NotImplemented, include = 'SKU')



__all__ = ('SKUEnhancement',)

from scarletio import RichAttributeErrorBaseType

from .fields import (
    parse_boost_cost, parse_guild, parse_purchase_limit,
    put_boost_cost, put_guild, put_purchase_limit,
    validate_boost_cost, validate_guild,
    validate_purchase_limit
)


class SKUEnhancement(RichAttributeErrorBaseType):
    """
    Represents an applied enhancements of an SKU.
    
    Attributes
    ----------
    boost_cost : `int`
        How much boost purchasing the SKU costs.
    
    guild : ``None | SKUEnhancementGuild``
        Guild specific enhancements granted by the SKU.
    
    purchase_limit : `int`
        The purchase limit of the SKU.
    """
    __slots__ = ('boost_cost', 'guild', 'purchase_limit')
    
    def __new__(
        cls,
        *,
        boost_cost = ...,
        guild = ...,
        purchase_limit = ...,
    ):
        """
        Creates a new user SKU enhancement instance from the given parameters.
        
        Attributes
        ----------
        boost_cost : `None | int`, Optional (Keyword only)
            How much boost purchasing the SKU costs.
        
        guild : ``None | SKUEnhancementGuild``, Optional (Keyword only)
            Guild specific enhancements granted by the SKU.
        
        purchase_limit : `None | int`, Optional (Keyword only)
            The purchase limit of the SKU.
        
        """
        # boost_cost
        if boost_cost is ...:
            boost_cost = 0
        else:
            boost_cost = validate_boost_cost(boost_cost)
        
        # guild
        if guild is ...:
            guild = None
        else:
            guild = validate_guild(guild)
        
        # purchase_limit
        if purchase_limit is ...:
            purchase_limit = 0
        else:
            purchase_limit = validate_purchase_limit(purchase_limit)
        
        # Construct
        self = object.__new__(cls)
        self.boost_cost = boost_cost
        self.guild = guild
        self.purchase_limit = purchase_limit
        return self
    
    
    def __repr__(self):
        """Returns repr(self)."""
        repr_parts = ['<', type(self).__name__]
        
        field_added = False
        
        # boost_cost
        boost_cost = self.boost_cost
        if boost_cost:
            if field_added:
                repr_parts.append(',')
            else:
                field_added = True
            
            repr_parts.append(' boost_cost = ')
            repr_parts.append(repr(boost_cost))
        
        # guild
        guild = self.guild
        if (guild is not None):
            if field_added:
                repr_parts.append(',')
            else:
                field_added = True
            
            repr_parts.append(' guild = ')
            repr_parts.append(repr(guild))
        
        # purchase_limit
        purchase_limit = self.purchase_limit
        if purchase_limit:
            if field_added:
                repr_parts.append(',')
            else:
                field_added = True
            
            repr_parts.append(' purchase_limit = ')
            repr_parts.append(repr(purchase_limit))
        
        repr_parts.append('>')
        return ''.join(repr_parts)
    
    
    def __hash__(self):
        """Returns hash(self)."""
        hash_value = 0
        
        # boost_cost
        hash_value ^ self.boost_cost << 4
        
        # guild
        guild = self.guild
        if (guild is not None):
            hash_value ^= hash(guild)
        
        # purchase_limit
        hash_value ^= self.purchase_limit << 12
        
        return hash_value
    
    
    def __eq__(self, other):
        """Returns self == other."""
        if type(self) is not type(other):
            return NotImplemented
        
        # boost_cost
        if self.boost_cost != other.boost_cost:
            return False
        
        # guild
        if self.guild != other.guild:
            return False
        
        # purchase_limit
        if self.purchase_limit != other.purchase_limit:
            return False
        
        return True
    
    
    @classmethod
    def from_data(cls, data):
        """
        Creates a SKU enhancement from the given data.
        
        Parameters
        ----------
        data : `dict<str, object>`
            Received SKU enhancement data.
        
        Returns
        -------
        self : `instance<cls>`
        """
        self = object.__new__(cls)
        self.boost_cost = parse_boost_cost(data)
        self.guild = parse_guild(data)
        self.purchase_limit = parse_purchase_limit(data)
        return self
    
    
    def to_data(self, *, defaults = False):
        """
        Serializes the SKU enhancement to a json serializable object.
        
        Parameters
        ----------
        defaults : `bool` = `False`, Optional (Keyword only)
            Whether default values should be included as well.
        
        Returns
        -------
        data : `dict<str, object>`
        """
        data = {}
        
        put_boost_cost(self.boost_cost, data, defaults)
        put_guild(self.guild, data, defaults)
        put_purchase_limit(self.purchase_limit, data, defaults)
        
        return data
    
    
    def copy(self):
        """
        Copies the SKU enhancement.
        
        Returns
        -------
        new : `instance<type<self>>`
        """
        new = object.__new__(type(self))
        new.boost_cost = self.boost_cost
        
        # guild
        guild = self.guild
        if (guild is not None):
            guild = guild.copy()
        new.guild = guild
        
        new.purchase_limit = self.purchase_limit
        return new
    
    
    def copy_with(
        self, 
        *,
        boost_cost = ...,
        guild = ...,
        purchase_limit = ...,
    ):
        """
        Copies the SKU enhancement with the given fields.
        
        Parameters
        ----------
        boost_cost : `None | int`, Optional (Keyword only)
            How much boost purchasing the SKU costs.
        
        guild : ``None | SKUEnhancementGuild``, Optional (Keyword only)
            Guild specific enhancements granted by the SKU.
        
        purchase_limit : `None | int`, Optional (Keyword only)
            The purchase limit of the SKU.
        
        Returns
        -------
        new : `instance<type<self>>`
        
        Raises
        ------
        TypeError
            - If a parameter's type is incorrect.
        ValueError
            - If a parameter's value is incorrect.
        """
        # boost_cost
        if boost_cost is ...:
            boost_cost = self.boost_cost
        else:
            boost_cost = validate_boost_cost(boost_cost)
        
        # guild
        if guild is ...:
            guild = self.guild
            if (guild is not None):
                guild = guild.copy()
        else:
            guild = validate_guild(guild)
        
        # purchase_limit
        if purchase_limit is ...:
            purchase_limit = self.purchase_limit
        else:
            purchase_limit = validate_purchase_limit(purchase_limit)
        
        # Construct
        new = object.__new__(type(self))
        new.boost_cost = boost_cost
        new.guild = guild
        new.purchase_limit = purchase_limit
        return new



__all__ = ('ChannelMetadataGuildMainBase',)

from scarletio import copy_docs

from ...core import GUILDS
from ...permission import Permission
from ...permission.permission import (
    PERMISSION_ALL, PERMISSION_MASK_ADMINISTRATOR, PERMISSION_MASK_VIEW_CHANNEL, PERMISSION_NONE
)
from ...user import ClientUserBase

from .fields import (
    parse_permission_overwrites, parse_position, put_permission_overwrites, put_position,
    validate_permission_overwrites, validate_position
)

from .guild_base import ChannelMetadataGuildBase


class ChannelMetadataGuildMainBase(ChannelMetadataGuildBase):
    """
    Base class for main guild metadata channels not including thread channels.
    
    Attributes
    ----------
    _cache_permission : ``None | dict<int, Permission>``
        A `user_id` to ``Permission`` relation mapping for caching permissions. Defaults to `None`.
    name : `str`
        The channel's name.
    parent_id : `int`
        The channel's parent's identifier.
    permission_overwrites : `None`, `dict` of (`int`, ``PermissionOverwrite``) items
        The channel's permission overwrites.
    position : `int`
        The channel's position.
    
    Class Attributes
    ----------------
    order_group: `int` = `0`
        The channel's order group used when sorting channels.
    """
    __slots__ = ('permission_overwrites', 'position', )
    
    
    def __new__(
        cls,
        *,
        name = ...,
        parent_id = ...,
        permission_overwrites = ...,
        position = ...,
    ):
        """
        Creates a new guild main base channel metadata from the given parameters.
        
        Parameters
        ----------
        name : `str`, Optional (Keyword only)
            The channel's name.
        parent_id : ``None | int | Channel``, Optional (Keyword only)
            The channel's parent's identifier.
        permission_overwrites : ``None | iterable<PermissionOverwrite>``, Optional (Keyword only)
            The channel's permission overwrites.
        position : `int`, Optional (Keyword only)
            The channel's position.
        
        Raises
        ------
        TypeError
            - If a parameter's type is incorrect.
        ValueError
            - If a parameter's value is incorrect.
        """
        # permission_overwrites
        if permission_overwrites is ...:
            permission_overwrites = None
        else:
            permission_overwrites = validate_permission_overwrites(permission_overwrites)
        
        # position
        if position is ...:
            position = 0
        else:
            position = validate_position(position)
        
        # Construct
        self = ChannelMetadataGuildBase.__new__(
            cls,
            name = name,
            parent_id = parent_id,
        )
        self.permission_overwrites = permission_overwrites
        self.position = position
        return self
    
    
    @classmethod
    @copy_docs(ChannelMetadataGuildBase.from_keyword_parameters)
    def from_keyword_parameters(cls, keyword_parameters):
        return cls(
            name = keyword_parameters.pop('name', ...),
            parent_id = keyword_parameters.pop('parent_id', ...),
            permission_overwrites = keyword_parameters.pop('permission_overwrites', ...),
            position = keyword_parameters.pop('position', ...),
        )
    
    
    @copy_docs(ChannelMetadataGuildBase.__hash__)
    def __hash__(self):
        hash_value = ChannelMetadataGuildBase.__hash__(self)
        
        # permission_overwrites
        permission_overwrites = self.permission_overwrites
        if (permission_overwrites is not None):
            hash_value ^= len(permission_overwrites) << 8
            
            for permission_overwrite in permission_overwrites.values():
                hash_value ^= hash(permission_overwrite)
        
        # position
        hash_value ^= self.position << 24
        
        return hash_value
    
    
    @copy_docs(ChannelMetadataGuildBase._created)
    def _created(self, channel_entity, client, strong_cache):
        if strong_cache:
            try:
                guild = GUILDS[channel_entity.guild_id]
            except KeyError:
                pass
            else:
                guild.channels[channel_entity.id] = channel_entity
    
    
    @copy_docs(ChannelMetadataGuildBase._delete)
    def _delete(self, channel_entity, client):
        try:
            guild = GUILDS[channel_entity.guild_id]
        except KeyError:
            pass
        else:
            try:
                del guild.channels[channel_entity.id]
            except KeyError:
                pass
    
    
    @copy_docs(ChannelMetadataGuildBase._is_equal_same_type)
    def _is_equal_same_type(self, other):
        if not ChannelMetadataGuildBase._is_equal_same_type(self, other):
            return False
        
        if self.permission_overwrites != other.permission_overwrites:
            return False
        
        if self.position != other.position:
            return False
        
        return True
    
    
    @copy_docs(ChannelMetadataGuildBase._update_attributes)
    def _update_attributes(self, data):
        ChannelMetadataGuildBase._update_attributes(self, data)
        
        # permission_overwrites
        self.permission_overwrites = parse_permission_overwrites(data)
        
        # position
        self.position = parse_position(data)
    
    
    @copy_docs(ChannelMetadataGuildBase._difference_update_attributes)
    def _difference_update_attributes(self, data):
        old_attributes = ChannelMetadataGuildBase._difference_update_attributes(self, data)
        
        # permission_overwrites
        permission_overwrites = parse_permission_overwrites(data)
        if self.permission_overwrites != permission_overwrites:
            old_attributes['permission_overwrites'] = self.permission_overwrites
            self.permission_overwrites = permission_overwrites
        
        
        # position
        position = parse_position(data)
        if self.position != position:
            old_attributes['position'] = self.position
            self.position = position
        
        
        return old_attributes
    
    
    @copy_docs(ChannelMetadataGuildBase.to_data)
    def to_data(self, *, defaults = False, include_internals = False):
        data = ChannelMetadataGuildBase.to_data(self, defaults = defaults, include_internals = include_internals)
        
        # position
        put_position(self.position, data, defaults)
        
        # permission_overwrites
        put_permission_overwrites(self.permission_overwrites, data, defaults)
        
        return data
    
    
    @classmethod
    @copy_docs(ChannelMetadataGuildBase._create_empty)
    def _create_empty(cls):
        self = super(ChannelMetadataGuildMainBase, cls)._create_empty()
        
        self.permission_overwrites = None
        self.position = 0
        
        return self
    
    
    @copy_docs(ChannelMetadataGuildBase.copy)
    def copy(self):
        new = ChannelMetadataGuildBase.copy(self)
        
        permission_overwrites = self.permission_overwrites
        if (permission_overwrites is not None):
            permission_overwrites = {
                target_id: permission_overwrite.copy()
                for target_id, permission_overwrite in permission_overwrites.items()
            }
        new.permission_overwrites = permission_overwrites
        
        new.position = self.position
        
        return new
    
    
    def copy_with(
        self,
        *,
        name = ...,
        parent_id = ...,
        permission_overwrites = ...,
        position = ...,
    ):
        """
        Copies the guild main base channel metadata with the given fields.
        
        Parameters
        ----------
        name : `str`, Optional (Keyword only)
            The channel's name.
        parent_id : ``None | int | Channel``, Optional (Keyword only)
            The channel's parent's identifier.
        permission_overwrites : ``None | iterable<PermissionOverwrite>``, Optional (Keyword only)
            The channel's permission overwrites.
        position : `int`, Optional (Keyword only)
            The channel's position.
        
        Returns
        -------
        new : `instance<type<self>>`
        
        Raises
        ------
        TypeError
            - If a parameter's type is incorrect.
        ValueError
            - If a parameter's value is incorrect.
        """
        # permission_overwrites
        if permission_overwrites is ...:
            permission_overwrites = self.permission_overwrites
            if (permission_overwrites is not None):
                permission_overwrites = {
                    target_id: permission_overwrite.copy()
                    for target_id, permission_overwrite in permission_overwrites.items()
                }
        else:
            permission_overwrites = validate_permission_overwrites(permission_overwrites)
        
        # position
        if position is ...:
            position = self.position
        else:
            position = validate_position(position)
        
        # Construct
        new = ChannelMetadataGuildBase.copy_with(
            self,
            name = name,
            parent_id = parent_id,
        )
        new.permission_overwrites = permission_overwrites
        new.position = position
        return new
    
    
    @copy_docs(ChannelMetadataGuildBase.copy_with_keyword_parameters)
    def copy_with_keyword_parameters(self, keyword_parameters):
        return self.copy_with(
            name = keyword_parameters.pop('name', ...),
            parent_id = keyword_parameters.pop('parent_id', ...),
            permission_overwrites = keyword_parameters.pop('permission_overwrites', ...),
            position = keyword_parameters.pop('position', ...),
        )
    
    
    def _get_base_permissions_for(self, channel_entity, user):
        """
        Base permission calculator method. Subclasses call this first, then apply their channel type related changes.
        
        Parameters
        ----------
        channel_entity : ``Channel``
            The channel entity owning the metadata.
        user : ``UserBase``
            The user to calculate it's permissions of.
        
        Returns
        -------
        permissions : ``Permission``
            The calculated permissions.
        """
        guild_id = channel_entity.guild_id
        try:
            guild = GUILDS[guild_id]
        except KeyError:
            return PERMISSION_NONE
        
        if user.id == guild.owner_id:
            return PERMISSION_ALL
        
        if not isinstance(user, ClientUserBase):
            if user.guild_id in guild.channels:
                role_everyone = guild.roles.get(guild_id, None)
                if role_everyone is None:
                    permissions = PERMISSION_NONE
                else:
                    permissions = role_everyone.permissions
                
                # Apply everyone's if applicable
                permission_overwrites = self.permission_overwrites
                if (permission_overwrites is not None):
                    try:
                        permission_overwrite_everyone = permission_overwrites[guild_id]
                    except KeyError:
                        pass
                    else:
                        permissions &= ~permission_overwrite_everyone.deny
                        permissions |= permission_overwrite_everyone.allow
            
                return permissions
            else:
                return PERMISSION_NONE
        
        # Are we in the guild?
        try:
            guild_profile = user.guild_profiles[guild_id]
        except KeyError:
            return PERMISSION_NONE
        
        # Apply everyone's
        role_everyone = guild.roles.get(guild_id, None)
        if role_everyone is None:
            permissions = 0
        else:
            permissions = role_everyone.permissions
        
        roles = guild_profile.roles
        
        # Apply permissions of each role of the user.
        if (roles is not None):
            for role in roles:
                permissions |= role.permissions
        
        permission_overwrites = self.permission_overwrites
        if (permission_overwrites is not None):
            # Apply everyone's if applicable
            try:
                permission_overwrite_everyone = permission_overwrites[guild_id]
            except KeyError:
                pass
            else:
                permissions &= ~permission_overwrite_everyone.deny
                permissions |= permission_overwrite_everyone.allow
            
            # Apply role overwrite
            allow = 0
            deny = 0
            
            if (roles is not None):
                for role in roles:
                    try:
                        permission_overwrite_role = permission_overwrites[role.id]
                    except KeyError:
                        pass
                    else:
                        allow |= permission_overwrite_role.allow
                        deny |= permission_overwrite_role.deny
            
            permissions &= ~deny
            permissions |= allow
            
            # Apply user specific
            try:
                permission_overwrite_user = permission_overwrites[user.id]
            except KeyError:
                pass
            else:
                permissions &= ~permission_overwrite_user.deny
                permissions |= permission_overwrite_user.allow
        
        # Are we admin?
        if permissions & PERMISSION_MASK_ADMINISTRATOR:
            return PERMISSION_ALL
        
        return Permission(permissions)
    
    
    @copy_docs(ChannelMetadataGuildBase._get_permissions_for)
    def _get_permissions_for(self, channel_entity, user):
        result = self._get_base_permissions_for(channel_entity, user)
        if not result & PERMISSION_MASK_VIEW_CHANNEL:
            result = PERMISSION_NONE
        
        return result
    
    
    def _get_base_permissions_for_roles(self, channel_entity, roles):
        """
        Returns the channel permissions of an imaginary user who would have the listed roles. This method is called
        first by subclasses to apply their own related permissions on it.
        
        Parameters
        ----------
        channel_entity : ``Channel``
            The channel entity owning the metadata.
        roles : ``tuple<Role>``
            The roles to calculate final permissions from.
        
        Returns
        -------
        permissions : ``Permission``
            The calculated permissions.
        
        Notes
        -----
        Partial roles and roles from other guilds as well are ignored.
        """
        guild_id = channel_entity.guild_id
        
        try:
            guild = GUILDS[guild_id]
        except KeyError:
            return PERMISSION_NONE
        
        role_everyone = guild.roles.get(guild_id, None)
        if role_everyone is None:
            permissions = 0
        else:
            permissions = role_everyone.permissions
        
        roles = sorted(roles)
        for role in roles:
            if role.guild_id == guild_id:
                permissions |= role.permissions
        
        if permissions & PERMISSION_MASK_ADMINISTRATOR:
            return PERMISSION_ALL
        
        permission_overwrites = self.permission_overwrites
        try:
            permission_overwrite_everyone = permission_overwrites[guild_id]
        except KeyError:
            pass
        else:
            permissions &= ~permission_overwrite_everyone.deny
            permissions |= permission_overwrite_everyone.allow
        
        allow = 0
        deny = 0
        
        for role in roles:
            try:
                permission_overwrite_role = permission_overwrites[role.id]
            except KeyError:
                pass
            else:
                allow |= permission_overwrite_role.allow
                deny |= permission_overwrite_role.deny
        
        permissions &= ~deny
        permissions |= allow
        
        return Permission(permissions)
    
    
    @copy_docs(ChannelMetadataGuildBase._get_permissions_for_roles)
    def _get_permissions_for_roles(self, channel_entity, roles):
        permissions = self._get_base_permissions_for_roles(channel_entity, roles)
        if not permissions & PERMISSION_MASK_VIEW_CHANNEL:
            permissions = PERMISSION_NONE
        
        return permissions



__all__ = ('PrivacyLevel', 'ScheduledEventEntityType', 'ScheduledEventStatus',)

from ...bases import Preinstance as P, PreinstancedBase

from ..scheduled_event_entity_metadata import (
    ScheduledEventEntityMetadataBase, ScheduledEventEntityMetadataLocation, ScheduledEventEntityMetadataStage
)


class ScheduledEventStatus(PreinstancedBase, value_type = int):
    """
    Represents a scheduled event's status.
    
    Attributes
    ----------
    name : `str`
        The name of the scheduled event status.
    
    value : `int`
        The identifier value the scheduled event status.
    
    Type Attributes
    ---------------
    Every predefined scheduled event can be accessed as type attribute as well:
    
    +-----------------------+---------------+-------+
    | Type attribute name   | Name          | Value |
    +=======================+===============+=======+
    | none                  | none          | 0     |
    +-----------------------+---------------+-------+
    | scheduled             | scheduled     | 1     |
    +-----------------------+---------------+-------+
    | active                | active        | 2     |
    +-----------------------+---------------+-------+
    | completed             | completed     | 3     |
    +-----------------------+---------------+-------+
    | cancelled             | cancelled     | 4     |
    +-----------------------+---------------+-------+
    """
    __slots__ = ()
    
    none = P(0, 'none')
    scheduled = P(1, 'scheduled')
    active = P(2, 'active')
    completed = P(3, 'completed')
    cancelled = P(4, 'cancelled')


class ScheduledEventEntityType(PreinstancedBase, value_type = int):
    """
    Represents a scheduled event's entity's type.
    
    Attributes
    ----------
    name : `str`
        The name of the scheduled event's entity's type.
    
    metadata_type : `type<ScheduledEventEntityMetadata>`
        The scheduled event's metadata's applicable type.
    
    value : `int`
        The identifier value the scheduled event's entity type.
    
    Type Attributes
    ---------------
    Every predefined scheduled event can be accessed as type attribute as well:
    
    +-----------------------+---------------+-------+---------------------------------------------+
    | Type attribute name   | Name          | Value | Metadata type                               |
    +=======================+===============+=======+=============================================+
    | none                  | none          | 0     | ``ScheduledEventEntityMetadataBase``        |
    +-----------------------+---------------+-------+---------------------------------------------+
    | stage                 | stage         | 1     | ``ScheduledEventEntityMetadataStage``       |
    +-----------------------+---------------+-------+---------------------------------------------+
    | voice                 | voice         | 2     | ``ScheduledEventEntityMetadataBase``        |
    +-----------------------+---------------+-------+---------------------------------------------+
    | location              | location      | 3     | ``ScheduledEventEntityMetadataLocation``    |
    +-----------------------+---------------+-------+---------------------------------------------+
    """
    __slots__ = ('metadata_type',)
    
    
    def __new__(cls, value, name = None, metadata_type = None):
        """
        Creates a new scheduled event entity type instance from the given parameters.
        
        Parameters
        ----------
        value : `int`
            The unique identifier of the scheduled event entity type.
        
        name : `None | str` = `None`, Optional
            The name of the scheduled event entity type.
        
        metadata_type : `None | type<ScheduledEventEntityMetadata>` = `None`, Optional
            The scheduled event's metadata's applicable type.
        """
        if metadata_type is None:
            metadata_type = ScheduledEventEntityMetadataBase
        
        self = PreinstancedBase.__new__(cls, value, name)
        self.metadata_type = metadata_type
        return self
    
    
    none = P(0, 'none', ScheduledEventEntityMetadataBase)
    stage = P(1, 'stage', ScheduledEventEntityMetadataStage)
    voice = P(2, 'voice', ScheduledEventEntityMetadataBase)
    location = P(3, 'location', ScheduledEventEntityMetadataLocation)


class PrivacyLevel(PreinstancedBase, value_type = int):
    """
    Represents a stage channel's privacy level.
    
    Attributes
    ----------
    name : `str`
        The name of the privacy level.
    
    value : `int`
        The identifier value the privacy level.
    
    Type Attributes
    ---------------
    Every predefined privacy level can be accessed as type attribute as well:
    
    +---------------------------+---------------------------+-------+
    | Type attribute name       | Name                      | Value |
    +===========================+===========================+=======+
    | none                      | none                      | 0     |
    +---------------------------+---------------------------+-------+
    | public                    | public                    | 1     |
    +---------------------------+---------------------------+-------+
    | guild_only                | guild only                | 2     |
    +---------------------------+---------------------------+-------+
    | public_with_join_request  | public with join request  | 3     |
    +---------------------------+---------------------------+-------+
    """
    __slots__ = ()
    
    none = P(0, 'none')
    public = P(1, 'public')
    guild_only = P(2, 'guild only')
    public_with_join_request = P(3, 'public with join request')



__all__ = ()

from ...message.message.fields import validate_channel_id, validate_id

from ..audit_log_entry_detail_conversion import AuditLogEntryDetailConversion, AuditLogEntryDetailConversionGroup
from ..conversion_helpers.converters import value_deserializer_id, value_serializer_id


# ---- channel_id ----

CHANNEL_ID_CONVERSION = AuditLogEntryDetailConversion(
    'channel_id',
    'channel_id',
    value_deserializer = value_deserializer_id,
    value_serializer = value_serializer_id,
    value_validator = validate_channel_id,
)


# ---- count ----

COUNT_CONVERSION = AuditLogEntryDetailConversion(
    'count',
    'count',
)

@COUNT_CONVERSION.set_value_deserializer
def count_deserializer(value):
    if value is None:
        value = 0
    return value


@COUNT_CONVERSION.set_value_validator
def validate_count(value):
    if not isinstance(value, int):
        raise TypeError(f'`count` can be `int`, got {value.__class__.__name__}; {value!r}.')
    
    if value < 0:
        raise ValueError(f'`count` cannot be negative, got {value!r}.')
    
    return value


# ---- id ----

ID_CONVERSION = AuditLogEntryDetailConversion(
    'message_id',
    'id',
    value_deserializer = value_deserializer_id,
    value_serializer = value_serializer_id,
    value_validator = validate_id,
)


# ---- Construct ----

MESSAGE_CONVERSIONS = AuditLogEntryDetailConversionGroup(
    CHANNEL_ID_CONVERSION,
    COUNT_CONVERSION,
    ID_CONVERSION,
)



__all__ = ('SubscriptionStatus',)

from ...bases import Preinstance as P, PreinstancedBase


class SubscriptionStatus(PreinstancedBase, value_type = int):
    """
    Represents a subscription's status.
    
    Attributes
    ----------
    name : `str`
        The name of the subscription status.
    
    value : `int`
        The identifier value the subscription status.
    
    Type Attributes
    ---------------
    Every predefined subscription status can be accessed as type attribute as well:
    
    +-----------------------+---------------+-------+
    | Type attribute name   | Name          | Value |
    +=======================+===============+=======+
    | active                | active        | 0     |
    +-----------------------+---------------+-------+
    | ending                | ending        | 1     |
    +-----------------------+---------------+-------+
    | inactive              | inactive      | 2     |
    +-----------------------+---------------+-------+
    """
    __slots__ = ()
    
    active = P(0, 'active')
    ending = P(1, 'ending')
    inactive = P(2, 'inactive')



__all__ = ('SlashCommandParameterAutoCompleter', )

from scarletio import RichAttributeErrorBaseType, copy_docs, export

from .....discord.client import Client
from .....discord.interaction import InteractionEvent

from ...constants import APPLICATION_COMMAND_FUNCTION_DEEPNESS
from ...converters import get_application_command_parameter_auto_completer_converters
from ...exceptions import handle_command_exception
from ...interfaces.command import CommandInterface
from ...interfaces.exception_handler import ExceptionHandlerInterface
from ...interfaces.self_reference import get_self_reference_of
from ...responding import process_command_coroutine
from ...utils import raw_name_to_display


@export
class SlashCommandParameterAutoCompleter(CommandInterface, ExceptionHandlerInterface, RichAttributeErrorBaseType):
    """
    Represents an application command parameter's auto completer.
    
    Attributes
    ----------
    _command_function : `async-callable˛
        The command's function to call.
    
    _exception_handlers : `None | list<CoroutineFunction>`
        Exception handlers added with ``.error`` to the interaction handler.
    
    _parameter_converters : `tuple<ParameterConverterBase>`
        Parsers to parse command parameters.
    
    _parent_reference : `None | WeakReferer<SelfReferenceInterface | object>`
        The parent slash command of the auto completer, where it was registered to.
    
    deepness : `int`
        How deep the auto completer was created. Deeper auto completers always overwrite higher ones.
    
    name_pairs : `frozenset<(str, str)>`
        Raw - display parameter names, to which the converter should autocomplete.
    """
    __slots__ = (
        '_command_function', '_exception_handlers', '_parameter_converters', '_parent_reference', 'deepness',
        'name_pairs'
    )
    
    def __new__(cls, function, parameter_names, deepness, parent):
        """
        Creates a new ``SlashCommandParameterAutoCompleter`` with the given parameters.
        
        Parameters
        ----------
        function : `async-callable`
            The function to create auto completer from.
        
        parameter_names : `list<str>`
            The names, which should be auto completed.
        
        deepness : `int`
            How deep the auto completer was created.
        
        parent : `None | SelfReferenceInterface | object`
            The parent of the auto completer.
        """
        command, parameter_converters = get_application_command_parameter_auto_completer_converters(function)
        
        name_pairs = frozenset((name, raw_name_to_display(name)) for name in parameter_names)
        
        parent_reference = get_self_reference_of(parent)
        
        self = object.__new__(cls)
        
        self._command_function = command
        self._exception_handlers = None
        self._parameter_converters = parameter_converters
        self._parent_reference = parent_reference
        self.deepness = deepness
        self.name_pairs = name_pairs
        
        return self
    

    def __repr__(self):
        """Returns the parameter auto completer's representation."""
        repr_parts = ['<', type(self).__name__]
        
        # name_pairs
        repr_parts.append(' name_pairs = ')
        repr_parts.append(repr(self.name_pairs))
        
        # _exception_handlers
        exception_handlers = self._exception_handlers
        if (exception_handlers is not None):
            repr_parts.append(', exception_handlers = ')
            repr_parts.append(repr(exception_handlers))
        
        # _command_function
        repr_parts.append(', command_function = ')
        repr_parts.append(repr(self._command_function))
        
        # deepness
        repr_parts.append(', deepness = ')
        repr_parts.append(repr(self.deepness))
        
        repr_parts.append('>')
        return ''.join(repr_parts)
    
    
    def __hash__(self):
        """Returns the parameter auto completer's hash value."""
        hash_value = 0
        
        # _command_function
        command_function = self._command_function
        try:
            command_function_hash_value = hash(command_function)
        except TypeError:
            command_function_hash_value = object.__hash__(command_function)
        hash_value ^= command_function_hash_value
        
        # _exception_handlers
        exception_handlers = self._exception_handlers
        if (exception_handlers is not None):
            hash_value ^= len(exception_handlers) << 4
            
            for exception_handler in exception_handlers:
                try:
                    exception_handler_hash_value = hash(exception_handler)
                except TypeError:
                    exception_handler_hash_value = object.__hash__(exception_handler)
                hash_value ^= exception_handler_hash_value
        
        # _parent_reference
        # Internal field
        
        # deepness
        hash_value ^= self.deepness
        
        # name_pairs
        hash_value ^= hash(self.name_pairs)
        
        return hash_value
    
    
    def __eq__(self, other):
        if type(self) is not type(other):
            return NotImplemented
        
        # _command_function
        if self._command_function != other._command_function:
            return False
        
        # _exception_handlers
        if self._exception_handlers != other._exception_handlers:
            return False
        
        # _parameter_converters
        # Internal field
        
        # _parent_reference
        # Internal field
        
        # deepness
        if self.deepness != other.deepness:
            return False
        
        # name_pairs
        if self.name_pairs != other.name_pairs:
            return False
        
        return True
    
    
    async def invoke(self, client, interaction_event):
        """
        Calls the parameter auto completer.
        
        This method is a coroutine.
        
        Parameters
        ----------
        client : ``Client``
            The respective client who received the event.
        interaction_event : ``InteractionEvent``
            The received interaction event.
        """
        parameters = []
        
        for parameter_converter in self._parameter_converters:
            parameter = await parameter_converter(client, interaction_event, None)
            parameters.append(parameter)
        
        auto_completer_coroutine = self._command_function(*parameters)
        
        try:
            await process_command_coroutine(
                client,
                interaction_event,
                None,
                auto_completer_coroutine,
            )
        except GeneratorExit:
            raise
        
        except BaseException as err:
            exception = err
        
        else:
            return
        
        # Do not put this into the `except` branch.
        try:
            await handle_command_exception(
                self,
                client,
                interaction_event,
                exception,
            )
        finally:
            exception = None
        return
    
    
    def copy(self):
        """
        Copies the parameter auto completer.
        
        Returns
        -------
        new : ``SlashCommandFunction``
        """
        new = object.__new__(type(self))
        
        # _command_function
        new._command_function = self._command_function
        
        # _exception_handlers
        exception_handlers = self._exception_handlers
        if (exception_handlers is not None):
            exception_handlers = exception_handlers.copy()
        new._exception_handlers = exception_handlers
        
        # _parameter_converters
        new._parameter_converters = self._parameter_converters
        
        # _parent_reference
        new._parent_reference = None
        
        # deepness
        new.deepness = self.deepness
        
        # name_pairs
        new.name_pairs = self.name_pairs
        
        return new
    
    
    def _bind_parent(self, new_parent):
        """
        Binds the parent application command function to self.
        
        If the auto completer is already bound to    an other object, will return a new one.
        
        Parameter
        ---------
        new_parent : `None`, ``SlashCommandFunction``
            The parent to bind to self.
        
        Returns
        -------
        new : ``SlashCommandParameterAutoCompleter``
            The new auto completer function bound to the new parent.
        """
        parent_reference = self._parent_reference
        if (parent_reference is None):
            self_parent = None
        else:
            self_parent = parent_reference()
        
        if (new_parent is None):
            if (self_parent is None):
                new = self
                new._parent_reference = None
            
            else:
                new = self.copy()
                new._parent_reference = None
        
        else:
            if (self_parent is None):
                new = self
                new._parent_reference = get_self_reference_of(new_parent)
            
            else:
                if (new_parent is self_parent):
                    new = self
                
                else:
                    new = self.copy()
                    new._parent_reference = get_self_reference_of(new_parent)
        
        return new
    
    
    def _is_deeper_than(self, other):
        """
        Returns whether self is deeper than other.
        
        Parameters
        ----------
        other : ``SlashCommandParameterAutoCompleter``
        """
        self_deepness = self.deepness
        other_deepness = other.deepness
        if self_deepness == APPLICATION_COMMAND_FUNCTION_DEEPNESS:
            return self_deepness != other_deepness
        
        if other_deepness == APPLICATION_COMMAND_FUNCTION_DEEPNESS:
            return False
        
        if self_deepness > other_deepness:
            return True
        
        return False
    
    
    def _difference_match_parameters(self, auto_completable_parameters):
        """
        Matches auto completable parameters returning a list of the matched ones.
        
        Parameters
        ----------
        auto_completable_parameters : `set<ParameterConverterSlashCommand>`
            Auto completable parameters.
        
        Returns
        -------
        matched : `list<ParameterConverterSlashCommand>`
            The matched parameters.
        """
        matched = []
        
        name_pairs = {*self.name_pairs}
        
        for name_pair in [*name_pairs]:
            name = name_pair[1]
            
            for parameter in auto_completable_parameters:
                if parameter.name == name:
                    name_pairs.discard(name_pair)
                    matched.append(parameter)
                    auto_completable_parameters.discard(parameter)
                    break
        
        for name_pair in [*name_pairs]:
            name = name_pair[0]
            
            for parameter in auto_completable_parameters:
                if parameter.parameter_name == name:
                    name_pairs.discard(name_pair)
                    matched.append(parameter)
                    auto_completable_parameters.discard(parameter)
                    break
        
        return matched
    
    
    @copy_docs(CommandInterface.get_command_function)
    def get_command_function(self):
        return self._command_function
